from .base import StandardFeature
from .standard_features import PeriodicSeasonalFeature, ConstantFeature, CustomFeature

__all__ = [
    "StandardFeature",
    "PeriodicSeasonalFeature",
    "ConstantFeature",
    "CustomFeature",
]



import numpy as np
import pandas as pd

from giottotime.feature_creation.index_dependent_features.tda_features import (
    AvgLifeTimeFeature,
)


def test_correct_average_lifetime_computation():
    np.random.seed(0)

    output_name = "ignored"
    avg_lifetime_feature = AvgLifeTimeFeature(output_name=output_name)
    df = pd.DataFrame(np.random.randint(0, 100, size=(30, 1)), columns=["ignored"])

    persistence_diagrams = avg_lifetime_feature._compute_persistence_diagrams(df)

    average_lifetime = avg_lifetime_feature._compute_average_lifetime(
        persistence_diagrams
    )

    expected_avg_lifetime = np.array(
        [
            0.8719302650092168,
            0.9375827753658988,
            0.9898459500730074,
            0.9927494731020051,
            0.9657270652747636,
            0.9135946743014184,
            0.838502504813531,
            0.8052276794850292,
            0.7657986690498374,
            0.8041319195886067,
            0.8091595180105875,
            0.8212692086741538,
            0.8097153100855028,
            0.8103762891856373,
            0.7947070443375637,
            0.8403939703821199,
            0.8371402400850239,
        ]
    )

    assert np.allclose(expected_avg_lifetime, average_lifetime)


def test_correct_avg_lifetime_feature():
    np.random.seed(0)

    output_name = "avg_lifetime_feature"
    avg_lifetime_feature = AvgLifeTimeFeature(output_name=output_name)
    df = pd.DataFrame(np.random.randint(0, 100, size=(30, 1)), columns=["old_name"])

    avg_lifetime = avg_lifetime_feature.fit_transform(df)

    expected_avg_lifetime = pd.DataFrame.from_dict(
        {
            output_name: [
                0.871930,
                0.871930,
                0.937583,
                0.937583,
                0.989846,
                0.989846,
                0.992749,
                0.992749,
                0.965727,
                0.965727,
                0.913595,
                0.913595,
                0.838503,
                0.838503,
                0.805228,
                0.805228,
                0.765799,
                0.765799,
                0.804132,
                0.804132,
                0.809160,
                0.809160,
                0.821269,
                0.821269,
                0.809715,
                0.809715,
                0.810376,
                0.794707,
                0.840394,
                0.837140,
            ]
        }
    )
    expected_avg_lifetime.index = df.index

    assert np.allclose(expected_avg_lifetime.values, avg_lifetime.values)
    assert np.array_equal(expected_avg_lifetime.index, avg_lifetime.index)
    assert expected_avg_lifetime.columns == avg_lifetime.columns



import numpy as np
import pandas as pd


class FeatureSplitter:
    """Splits the feature matrices X and y in X_train, y_train, X_test, y_test.

    X and y are the feature matrices obtained from the FeatureCreation class.

    Parameters
    ----------
    drop_na_mode: str, optional, default: ``'any'``
        How to drop the Nan contained in the ``X`` and ``y`` matrices. Only 'any' is
        supported for the moment.

    Examples
    --------
    >>> import pandas as pd
    >>> import numpy as np
    >>> from giottotime.model_selection import FeatureSplitter
    >>> X = pd.DataFrame.from_dict({"feature_0": [np.nan, 0, 1, 2, 3, 4, 5, 6, 7, 8],
    ...                             "feature_1": [np.nan, np.nan, 0.5, 1.5, 2.5, 3.5,
    ...                                            4.5, 5.5, 6.5, 7.5, ]
    ...                            })
    >>> y = pd.DataFrame.from_dict({"y_0": [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
    ...                             "y_1": [1, 2, 3, 4, 5, 6, 7, 8, 9, np.nan],
    ...                             "y_2": [2, 3, 4, 5, 6, 7, 8, 9, np.nan, np.nan]
    ...                            })
    >>> feature_splitter = FeatureSplitter()
    >>> X_train, y_train, X_test, y_test = feature_splitter.transform(X, y)
    >>> X_train
           feature_0            feature_1
    2           1.0                   0.5
    3           2.0                   1.5
    4           3.0                   2.5
    5           4.0                   3.5
    6           5.0                   4.5
    7           6.0                   5.5
    >>> y_train
       y_0  y_1  y_2
    2    2  3.0  4.0
    3    3  4.0  5.0
    4    4  5.0  6.0
    5    5  6.0  7.0
    6    6  7.0  8.0
    7    7  8.0  9.0
    >>> X_test
          feature_0             feature_1
    8           7.0                   6.5
    9           8.0                   7.5
    >>> y_test
       y_0  y_1  y_2
    8    8  9.0  NaN
    9    9  NaN  NaN

    """

    def __init__(self, drop_na_mode: str = "any"):
        if drop_na_mode != "any":
            raise ValueError(
                f'Only drop_na_mode="any" is supported. Detected: {drop_na_mode}'
            )
        self.drop_na_mode = drop_na_mode

    def transform(
        self, X: pd.DataFrame, y: pd.DataFrame
    ) -> (pd.DataFrame, pd.DataFrame, pd.DataFrame, pd.DataFrame):
        """Split the feature matrices X and y in X_train, y_train, X_test, y_test.

        ``X`` and ``y`` are the feature matrices obtained from the FeatureCreation
        class.

        Parameters
        ----------
        X : pd.DataFrame, shape (n_samples, n_features), required
            The feature matrix.

        y : ``pd.DataFrame``, shape (n_samples, horizon), required
            The y matrix.

        Returns
        -------
        X_train, y_train, X_test, y_test : ``Tuple[pd.DataFrame, pd.DataFrame, \
            pd.DataFrame, pd.DataFrame]``
            The X and y, split between train and test.

        """
        X, y = self._drop_X_na(X, y)
        X_train, y_train, X_test, y_test = self._split_train_test(X, y)
        return X_train, y_train, X_test, y_test

    def _drop_X_na(
        self, X: pd.DataFrame, y: pd.DataFrame
    ) -> (pd.DataFrame, pd.DataFrame):

        X = X.dropna(axis=0, how=self.drop_na_mode)
        y = y.loc[X.index]
        return X, y

    def _split_train_test(
        self, X: pd.DataFrame, y: pd.DataFrame
    ) -> (pd.DataFrame, pd.DataFrame, pd.DataFrame, pd.DataFrame):

        train_indexes, test_indexes = self._get_train_test_indexes_from_y(y)
        X_train, y_train = X.loc[train_indexes], y.loc[train_indexes]
        X_test, y_test = X.loc[test_indexes], y.loc[test_indexes]
        return X_train, y_train, X_test, y_test

    def _get_train_test_indexes_from_y(self, y):
        last_train_index = self._last_non_nan_y_index(y)
        train_indexes = y.loc[:last_train_index].index if last_train_index else []
        test_indexes = y.index.difference(train_indexes)
        return train_indexes, test_indexes

    def _last_non_nan_y_index(self, y: pd.DataFrame) -> pd.Period:
        y_nan = y.isnull().any(axis=1).replace(True, np.nan)
        return y_nan.last_valid_index()



import numpy as np
import pandas as pd
import pytest
from hypothesis import given, settings

from giottotime.feature_creation import CalendarFeature
from giottotime.utils.hypothesis.time_indexes import giotto_time_series


def test_empty_and_non_finite_kernel_error():
    with pytest.raises(ValueError):
        CalendarFeature(
            start_date="ignored",
            end_date="ignored",
            region="america",
            country="Brazil",
            kernel=np.array([]),
            output_name="cal",
        )

    with pytest.raises(ValueError):
        CalendarFeature(
            start_date="ignored",
            end_date="ignored",
            region="america",
            country="Brazil",
            kernel=np.array([np.nan, 1]),
            output_name="cal",
        )


def test_unevenly_spaced_time_series():
    unevenly_spaced_ts = pd.DataFrame(
        index=[
            pd.Period("2012-01-01"),
            pd.Period("2012-01-03"),
            pd.Period("2012-01-10"),
        ]
    )
    cal_feature = CalendarFeature(
        start_date="ignored",
        end_date="ignored",
        region="america",
        country="Brazil",
        kernel=np.array([0, 1]),
        output_name="cal",
    )

    with pytest.raises(ValueError):
        cal_feature.transform(unevenly_spaced_ts)


def test_correct_calendar_without_input_ts():
    start_date = "2018-01-01"
    end_date = "2019-01-01"

    cal_feature = CalendarFeature(
        start_date=start_date,
        end_date=end_date,
        region="america",
        country="Brazil",
        kernel=np.array([0, 1]),
        output_name="cal",
    )
    Xt = cal_feature.transform()
    expected_index = pd.date_range(start_date, end_date)
    np.testing.assert_array_equal(Xt.index, expected_index)


@settings(deadline=pd.Timedelta(milliseconds=5000), max_examples=7)
@given(giotto_time_series(min_length=2, max_length=30))
def test_correct_index_random_ts(ts):
    output_name = "cal"
    cal_feature = CalendarFeature(
        start_date="ignored",
        end_date="ignored",
        region="america",
        country="Brazil",
        kernel=np.array([1, 2]),
        output_name=output_name,
    )
    Xt = cal_feature.transform(ts)
    np.testing.assert_array_equal(Xt.index, ts.index)



import numpy as np
import pandas as pd
import pandas.util.testing as testing

from giottotime.feature_creation import MovingAverageFeature
from giottotime.feature_creation.base import Feature


class BaseFeature(Feature):
    def __init__(self, output_name):
        super().__init__(output_name=output_name)

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        pass


def test_correct_renaming_single_col():
    n_cols = 1
    testing.N, testing.K = 500, n_cols
    df = testing.makeTimeDataFrame(freq="MS")

    output_name = "shift"
    shift_feature = BaseFeature(output_name=output_name)
    df_renamed = shift_feature._rename_columns(df)

    assert df.shape == df_renamed.shape

    expected_cols = output_name

    np.testing.assert_array_equal(expected_cols, df_renamed.columns)


def test_correct_renaming_series():
    n_cols = 1
    testing.N, testing.K = 500, n_cols
    df = testing.makeTimeSeries(freq="MS")

    output_name = "shift"
    shift_feature = BaseFeature(output_name=output_name)
    df_renamed = shift_feature._rename_columns(df)

    assert df.shape[0] == df_renamed.shape[0]
    assert pd.DataFrame == type(df_renamed)

    expected_cols = output_name

    np.testing.assert_array_equal(expected_cols, df_renamed.columns)


def test_correct_renaming_multiple_columns():
    n_cols = 10
    testing.N, testing.K = 500, n_cols
    df = testing.makeTimeDataFrame(freq="MS")

    output_name = "shift"
    shift_feature = BaseFeature(output_name=output_name)
    df_renamed = shift_feature._rename_columns(df)

    assert df.shape == df_renamed.shape

    expected_cols = [f"{output_name}_{k}" for k in range(n_cols)]

    np.testing.assert_array_equal(expected_cols, df_renamed.columns)


def test_correct_fit_transform():
    n_cols = 10
    testing.N, testing.K = 500, n_cols
    df = testing.makeTimeDataFrame(freq="MS")

    output_name = "shift"
    ma_feature = MovingAverageFeature(window_size=2, output_name=output_name)

    fit_transform_res = ma_feature.fit_transform(df)

    transform_res = ma_feature.fit(df).transform(df)

    testing.assert_frame_equal(fit_transform_res, transform_res)



from giottotime.feature_creation import (
    DetrendedFeature,
    RemovePolynomialTrend,
    RemoveExponentialTrend,
)
import pandas as pd

from giottotime.models import PolynomialTrend


def test_correct_index_features():
    output_name = "detrended_feature"
    time_index = pd.date_range(start="2020-01-01", end="2020-01-20")
    ts = pd.DataFrame(range(0, 20), index=time_index)
    model = PolynomialTrend()

    detrend_feature = DetrendedFeature(trend_model=model, output_name=output_name)
    ts_t = detrend_feature.transform(ts)
    expected_ts = pd.DataFrame(
        [
            1.22681324e-05,
            8.34525141e-06,
            4.86108426e-06,
            1.81563099e-06,
            -7.91108403e-07,
            -2.95913392e-06,
            -4.68844555e-06,
            -5.97904330e-06,
            -6.83092717e-06,
            -7.24409716e-06,
            -7.21855327e-06,
            -6.75429551e-06,
            -5.85132385e-06,
            -4.50963832e-06,
            -2.72923891e-06,
            -5.10125625e-07,
            2.14770155e-06,
            5.24424260e-06,
            8.77949753e-06,
            1.27534663e-05,
        ],
        columns=[output_name],
        index=time_index,
    )
    pd.testing.assert_frame_equal(ts_t, expected_ts)


def test_correct_polynomial_trend():
    output_name = "detrended_pol_feature"
    time_index = pd.date_range(start="2020-01-01", end="2020-01-20")
    ts = pd.DataFrame(range(0, 20), index=time_index)

    detrend_feature = RemovePolynomialTrend(polynomial_order=3, output_name=output_name)
    ts_t = detrend_feature.transform(ts)
    expected_ts = pd.DataFrame(
        [
            -6.59832896e-04,
            -2.39271025e-04,
            6.38171382e-05,
            2.63303644e-04,
            3.73060540e-04,
            4.06959876e-04,
            3.78873701e-04,
            3.02674064e-04,
            1.92233012e-04,
            6.14225966e-05,
            -7.58851350e-05,
            -2.05818134e-04,
            -3.14504350e-04,
            -3.88071736e-04,
            -4.12648241e-04,
            -3.74361819e-04,
            -2.59340418e-04,
            -5.37119912e-05,
            2.56395511e-04,
            6.84854138e-04,
        ],
        columns=[output_name],
        index=time_index,
    )
    pd.testing.assert_frame_equal(ts_t, expected_ts)


def test_correct_exponential_trend():
    output_name = "detrended_exp_feature"
    time_index = pd.date_range(start="2020-01-01", end="2020-01-20")
    ts = pd.DataFrame(range(0, 20), index=time_index)

    detrend_feature = RemoveExponentialTrend(output_name=output_name)
    ts_t = detrend_feature.transform(ts)
    expected_ts = pd.DataFrame(
        [
            -1.0,
            -0.18238542,
            0.60196471,
            1.34698345,
            2.04549733,
            2.68902453,
            3.26753629,
            3.76917473,
            4.1799193,
            4.48319226,
            4.65939237,
            4.68534338,
            4.53364205,
            4.17188719,
            3.5617681,
            2.65798675,
            1.40698343,
            -0.25457009,
            -2.40155216,
            -5.1224979,
        ],
        columns=[output_name],
        index=time_index,
    )
    pd.testing.assert_frame_equal(ts_t, expected_ts)



from hypothesis import given, strategies as st, settings
import pandas as pd
import pytest
import numpy as np
from hypothesis._settings import duration

from giottotime.utils.hypothesis.time_indexes import giotto_time_series
from giottotime.feature_creation import ShiftFeature, MovingAverageFeature
from giottotime.feature_creation.feature_creation import (
    _check_feature_names,
    FeatureCreation,
)


def test_wrong_feature_naming():
    features = [ShiftFeature(k, output_name=f"{k}") for k in range(50)]
    features_same_name = ShiftFeature(shift=100, output_name="1")
    features.append(features_same_name)

    with pytest.raises(ValueError):
        _check_feature_names(features)


def test_correct_feature_names():
    features = [ShiftFeature(k, output_name=f"{k}") for k in range(50)]
    _check_feature_names(features)


class TestFeatureCreation:
    def _correct_x(self, ts, time_series_features):
        features = pd.DataFrame(index=ts.index)
        for time_series_feature in time_series_features:
            x_transformed = time_series_feature.fit_transform(ts)
            features = pd.concat([features, x_transformed], axis=1)
        return features

    def _correct_y(self, ts, horizon):
        y = pd.DataFrame(index=ts.index)
        for k in range(1, horizon + 1):
            shift_feature = ShiftFeature(-k, f"shift_{k}")
            y[f"y_{k}"] = shift_feature.fit_transform(ts)
        return y

    def test_wrong_feature_naming(self):
        time_series_features = [
            ShiftFeature(shift=2, output_name="same_name"),
            ShiftFeature(shift=5, output_name="same_name"),
        ]

        with pytest.raises(ValueError):
            FeatureCreation(horizon=1, time_series_features=time_series_features)

    def test_correct_y_shifts(self):
        horizon = 2
        ts = pd.DataFrame.from_dict({"ignored": range(10)})
        ts.index = np.random.random(len(ts))

        feature_creation = FeatureCreation(horizon=horizon, time_series_features=[])
        y_shifts = feature_creation._create_y_shifts(ts)
        expected_y_shifts = pd.DataFrame.from_dict(
            {
                "y_1": [1, 2, 3, 4, 5, 6, 7, 8, 9, np.nan],
                "y_2": [2, 3, 4, 5, 6, 7, 8, 9, np.nan, np.nan],
            }
        )
        expected_y_shifts.index = ts.index

        pd.testing.assert_frame_equal(expected_y_shifts, y_shifts)

    @given(giotto_time_series(), st.integers(1, 10))
    def test_correct_y_shifts_random_ts(self, ts, horizon):
        feature_creation = FeatureCreation(horizon=horizon, time_series_features=[])
        y_shifts = feature_creation._create_y_shifts(ts)
        expected_y_shifts = self._correct_y(ts, horizon)

        pd.testing.assert_frame_equal(expected_y_shifts, y_shifts)

    def test_correct_x_shifts(self):
        horizon = 2
        ts = pd.DataFrame.from_dict({"ignored": range(10)})
        ts.index = np.random.random(len(ts))
        features = [
            MovingAverageFeature(window_size=2, output_name="mov_avg_2"),
            MovingAverageFeature(window_size=5, output_name="mov_avg_5"),
            ShiftFeature(shift=3, output_name="shift_3"),
        ]

        feature_creation = FeatureCreation(
            horizon=horizon, time_series_features=features
        )
        x_shifts = feature_creation._create_x_features(ts)
        expected_x_shifts = pd.DataFrame.from_dict(
            {
                "mov_avg_2": [np.nan, 0.5, 1.5, 2.5, 3.5, 4.5, 5.5, 6.5, 7.5, 8.5,],
                "mov_avg_5": [np.nan, np.nan, np.nan, np.nan, 2, 3, 4, 5, 6, 7,],
                "shift_3": [np.nan, np.nan, np.nan, 0, 1, 2, 3, 4, 5, 6,],
            }
        )
        expected_x_shifts.index = ts.index

        pd.testing.assert_frame_equal(expected_x_shifts, x_shifts)

    @given(giotto_time_series(), st.integers(1, 10))
    def test_correct_y_shifts_random_ts(self, ts, horizon):
        features = [
            MovingAverageFeature(window_size=2, output_name="mov_avg_2"),
            MovingAverageFeature(window_size=5, output_name="mov_avg_5"),
            ShiftFeature(shift=3, output_name="shift_3"),
        ]

        feature_creation = FeatureCreation(
            horizon=horizon, time_series_features=features
        )
        x_shifts = feature_creation._create_x_features(ts)
        expected_x_shifts = self._correct_x(ts, features)

        pd.testing.assert_frame_equal(expected_x_shifts, x_shifts)

    def test_correct_fit_transform(self):
        np.random.seed(0)

        random_index = np.random.random(10)
        horizon = 3
        ts = pd.DataFrame.from_dict({"ignored": range(10)})
        ts.index = random_index

        time_series_features = [
            ShiftFeature(shift=2, output_name="shift_2"),
            ShiftFeature(shift=-5, output_name="shift_5"),
        ]

        fc = FeatureCreation(horizon=horizon, time_series_features=time_series_features)
        x, y = fc.fit_transform(ts)

        expected_x = pd.DataFrame.from_dict(
            {
                "shift_2": [np.nan, np.nan, 0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0,],
                "shift_5": [
                    5.0,
                    6.0,
                    7.0,
                    8.0,
                    9.0,
                    np.nan,
                    np.nan,
                    np.nan,
                    np.nan,
                    np.nan,
                ],
            }
        )
        expected_x.index = random_index
        pd.testing.assert_frame_equal(expected_x, x)

        expected_y = pd.DataFrame.from_dict(
            {
                "y_1": [1, 2, 3, 4, 5, 6, 7, 8, 9, np.nan,],
                "y_2": [2, 3, 4, 5, 6, 7, 8, 9, np.nan, np.nan,],
                "y_3": [3, 4, 5, 6, 7, 8, 9, np.nan, np.nan, np.nan,],
            }
        )
        expected_y.index = random_index
        pd.testing.assert_frame_equal(expected_y, y)

    @given(giotto_time_series(min_length=1), st.integers(1, 10))
    def test_correct_fit_random_ts(self, ts, horizon):
        time_series_features = [
            ShiftFeature(shift=2, output_name="shift_2"),
            ShiftFeature(shift=-5, output_name="shift_5"),
        ]

        fc = FeatureCreation(horizon=horizon, time_series_features=time_series_features)
        x, y = fc.fit_transform(ts)

        expected_x = self._correct_x(ts, time_series_features)
        pd.testing.assert_frame_equal(expected_x, x)

        expected_y = self._correct_y(ts, horizon)
        pd.testing.assert_frame_equal(expected_y, y)



"""
The :mod:`giottotime.feature_creation` module contains a collection of different kind
of machine learning models, for dealing with time series and non time series data.
"""

from .regressors import LinearRegressor
from .time_series_models import GAR
from .trend_models import PolynomialTrend, ExponentialTrend

__all__ = [
    "LinearRegressor",
    "GAR",
    "PolynomialTrend",
    "ExponentialTrend",
]



from typing import Tuple

import hypothesis.strategies as st
import numpy as np
import pandas as pd
import pytest
from pandas.testing import assert_series_equal, assert_frame_equal
from hypothesis.extra.numpy import arrays
from hypothesis import given, settings, HealthCheck

from giottotime.utils.hypothesis.time_indexes import (
    series_with_timedelta_index,
    series_with_period_index,
    series_with_datetime_index,
    available_freqs,
)
from giottotime.time_series_preparation import TimeSeriesPreparation
from .utils import (
    pandas_series_with_period_index,
    datetime_index_series_to_period_index_series,
    timedelta_index_series_to_period_index_series,
)


class TestConstructor:
    @given(st.datetimes(), available_freqs(), st.booleans(), st.text())
    def test_constructor_does_not_fail(
        self,
        start: pd.datetime,
        freq: pd.Timedelta,
        resample_if_not_equispaced: bool,
        output_name: str,
    ):
        TimeSeriesPreparation(
            start=start,
            freq=freq,
            resample_if_not_equispaced=resample_if_not_equispaced,
            output_name=output_name,
        )

    @given(st.datetimes(), available_freqs(), st.booleans(), st.text())
    def test_constructor_initializes_parameters(
        self,
        start: pd.datetime,
        freq: pd.Timedelta,
        resample_if_not_equispaced: bool,
        output_name: str,
    ):
        time_series_preparation = TimeSeriesPreparation(
            start=start,
            freq=freq,
            resample_if_not_equispaced=resample_if_not_equispaced,
            output_name=output_name,
        )

        assert time_series_preparation.start == start
        assert time_series_preparation.end is None
        assert time_series_preparation.freq == freq
        assert (
            time_series_preparation.resample_if_not_equispaced
            == resample_if_not_equispaced
        )
        assert time_series_preparation.output_name == output_name


class TestToTimeIndexSeries:
    @given(st.lists(st.floats()), st.datetimes(), available_freqs())
    def test_list_as_input(
        self, input_list: pd.Series, start: pd.datetime, freq: pd.Timedelta,
    ):
        time_series_preparation = TimeSeriesPreparation(start=start, freq=freq)
        computed_time_series = time_series_preparation._to_time_index_series(input_list)
        expected_time_series = pandas_series_with_period_index(
            input_list, start, freq=freq
        )
        assert_series_equal(computed_time_series, expected_time_series)

    @given(
        arrays(shape=st.integers(0, 1000), dtype=float),
        st.datetimes(),
        available_freqs(),
    )
    def test_array_as_input(
        self, input_array: np.ndarray, start: pd.datetime, freq: pd.Timedelta,
    ):
        time_series_preparation = TimeSeriesPreparation(start=start, freq=freq)
        computed_time_series = time_series_preparation._to_time_index_series(
            input_array
        )
        expected_time_series = pandas_series_with_period_index(
            input_array, start, freq=freq
        )
        assert_series_equal(computed_time_series, expected_time_series)

    @given(series_with_period_index(), st.datetimes(), available_freqs())
    def test_period_index_series_unchanged(
        self, period_index_series: pd.Series, start: pd.datetime, freq: pd.Timedelta,
    ):
        time_series_preparation = TimeSeriesPreparation(start=start, freq=freq)
        computed_time_series = time_series_preparation._to_time_index_series(
            period_index_series
        )
        assert_series_equal(computed_time_series, period_index_series)

    @given(series_with_datetime_index(), st.datetimes(), available_freqs())
    def test_datetime_index_series_unchanged(
        self, datetime_index_series: pd.Series, start: pd.datetime, freq: pd.Timedelta,
    ):
        time_series_preparation = TimeSeriesPreparation(start=start, freq=freq)
        computed_time_series = time_series_preparation._to_time_index_series(
            datetime_index_series
        )
        assert_series_equal(computed_time_series, datetime_index_series)

    @given(series_with_timedelta_index(), st.datetimes(), available_freqs())
    def test_timedelta_index_series_unchanged(
        self, timedelta_index_series: pd.Series, start: pd.datetime, freq: pd.Timedelta,
    ):
        time_series_preparation = TimeSeriesPreparation(start=start, freq=freq)
        computed_time_series = time_series_preparation._to_time_index_series(
            timedelta_index_series
        )
        assert_series_equal(computed_time_series, timedelta_index_series)

    @given(st.tuples())
    def test_wrong_input_type(self, wrong_input: Tuple):
        time_series_preparation = TimeSeriesPreparation()
        with pytest.raises(TypeError):
            time_series_preparation._to_time_index_series(wrong_input)

    @given(series_with_period_index(), st.datetimes(), available_freqs())
    def test_period_index_dataframe_unchanged(
        self, period_index_series: pd.Series, start: pd.datetime, freq: pd.Timedelta,
    ):
        period_index_dataframe = pd.DataFrame(period_index_series)
        time_series_preparation = TimeSeriesPreparation(start=start, freq=freq)
        computed_time_series = time_series_preparation._to_time_index_series(
            period_index_dataframe
        )
        assert_series_equal(computed_time_series, period_index_series)

    @given(series_with_datetime_index(), st.datetimes(), available_freqs())
    def test_datetime_index_dataframe_unchanged(
        self, datetime_index_series: pd.Series, start: pd.datetime, freq: pd.Timedelta,
    ):
        datetime_index_dataframe = pd.DataFrame(datetime_index_series)
        time_series_preparation = TimeSeriesPreparation(start=start, freq=freq)
        computed_time_series = time_series_preparation._to_time_index_series(
            datetime_index_dataframe
        )
        assert_series_equal(computed_time_series, datetime_index_series)

    @given(series_with_timedelta_index(), st.datetimes(), available_freqs())
    def test_timedelta_index_dataframe_unchanged(
        self, timedelta_index_series: pd.Series, start: pd.datetime, freq: pd.Timedelta,
    ):
        timedelta_index_dataframe = pd.DataFrame(timedelta_index_series)
        time_series_preparation = TimeSeriesPreparation(start=start, freq=freq)
        computed_time_series = time_series_preparation._to_time_index_series(
            timedelta_index_dataframe
        )
        assert_series_equal(computed_time_series, timedelta_index_series)


class TestToEquispacedTimeSeries:
    @given(
        series_with_period_index(), st.datetimes(), available_freqs(), st.text(),
    )
    def test_with_resample_false(
        self,
        series: pd.Series,
        start: pd.datetime,
        freq: pd.Timedelta,
        output_name: str,
    ):
        time_series_preparation = TimeSeriesPreparation(
            start=start,
            freq=freq,
            resample_if_not_equispaced=False,
            output_name=output_name,
        )
        computed_series = time_series_preparation._to_equispaced_time_series(series)
        assert_series_equal(computed_series, series)

    @given(series_with_period_index(), st.datetimes(), available_freqs(), st.text())
    def test_with_resample_true(
        self,
        series: pd.Series,
        start: pd.datetime,
        freq: pd.Timedelta,
        output_name: str,
    ):
        time_series_preparation = TimeSeriesPreparation(
            start=start,
            freq=freq,
            resample_if_not_equispaced=True,
            output_name=output_name,
        )
        with pytest.raises(NotImplementedError):
            time_series_preparation._to_equispaced_time_series(series)


class TestToPeriodIndex:
    @settings(suppress_health_check=[HealthCheck.filter_too_much])
    @given(series_with_period_index())
    def test_with_period_index_as_input(self, period_index_series):
        time_series_preparation = TimeSeriesPreparation()
        computed_time_series = time_series_preparation._to_period_index_time_series(
            period_index_series
        )
        assert_series_equal(computed_time_series, period_index_series)

    @given(series_with_datetime_index())
    def test_with_datetime_index_as_input(self, datetime_index_series):
        time_series_preparation = TimeSeriesPreparation()
        computed_time_series = time_series_preparation._to_period_index_time_series(
            datetime_index_series
        )
        expected_time_series = datetime_index_series_to_period_index_series(
            computed_time_series
        )
        assert_series_equal(computed_time_series, expected_time_series)

    @given(series_with_timedelta_index())
    def test_with_timedelta_index_as_input(self, timedelta_index_series):
        time_series_preparation = TimeSeriesPreparation()
        computed_time_series = time_series_preparation._to_period_index_time_series(
            timedelta_index_series
        )
        expected_time_series = timedelta_index_series_to_period_index_series(
            timedelta_index_series
        )
        assert_series_equal(computed_time_series, expected_time_series)


class TestToPeriodIndexDataFrame:
    @given(series_with_period_index(), st.text())
    def test_output_dataframe_is_correct(
        self, period_index_series: pd.Series, output_name: str
    ):
        time_series_preparation = TimeSeriesPreparation(output_name=output_name)
        computed_time_series = time_series_preparation._to_period_index_dataframe(
            period_index_series
        )
        expected_time_series = pd.DataFrame({output_name: period_index_series})
        assert_frame_equal(computed_time_series, expected_time_series)


class TestTransform:
    @given(st.lists(st.floats()), st.datetimes(), available_freqs(), st.text())
    def test_list_as_input(
        self,
        input_list: pd.Series,
        start: pd.datetime,
        freq: pd.Timedelta,
        output_name: str,
    ):
        time_series_preparation = TimeSeriesPreparation(
            start=start, freq=freq, output_name=output_name
        )
        computed_time_series = time_series_preparation.transform(input_list)
        expected_series = pandas_series_with_period_index(input_list, start, freq=freq)
        expected_time_series = pd.DataFrame({output_name: expected_series})
        assert_frame_equal(computed_time_series, expected_time_series)

    @given(
        arrays(shape=st.integers(0, 1000), dtype=float),
        st.datetimes(),
        available_freqs(),
        st.text(),
    )
    def test_array_as_input(
        self,
        input_array: np.ndarray,
        start: pd.datetime,
        freq: pd.Timedelta,
        output_name: str,
    ):
        time_series_preparation = TimeSeriesPreparation(
            start=start, freq=freq, output_name=output_name
        )
        computed_time_series = time_series_preparation.transform(input_array)
        expected_series = pandas_series_with_period_index(input_array, start, freq=freq)
        expected_time_series = pd.DataFrame({output_name: expected_series})
        assert_frame_equal(computed_time_series, expected_time_series)

    @given(series_with_period_index(), st.datetimes(), available_freqs(), st.text())
    def test_period_index_as_input(
        self,
        period_index_series: pd.Series,
        start: pd.datetime,
        freq: pd.Timedelta,
        output_name: str,
    ):
        time_series_preparation = TimeSeriesPreparation(
            start=start, freq=freq, output_name=output_name
        )
        computed_time_series = time_series_preparation.transform(period_index_series)
        expected_time_series = pd.DataFrame({output_name: period_index_series})
        assert_frame_equal(computed_time_series, expected_time_series)

    @given(series_with_datetime_index(), st.datetimes(), available_freqs(), st.text())
    def test_datetime_index_as_input(
        self,
        datetime_index_series: pd.Series,
        start: pd.datetime,
        freq: pd.Timedelta,
        output_name: str,
    ):
        time_series_preparation = TimeSeriesPreparation(
            start=start, freq=freq, output_name=output_name
        )
        computed_time_series = time_series_preparation.transform(datetime_index_series)
        expected_series = datetime_index_series_to_period_index_series(
            datetime_index_series, freq=freq
        )
        expected_time_series = pd.DataFrame({output_name: expected_series})
        assert_frame_equal(computed_time_series, expected_time_series)

    @given(series_with_timedelta_index(), st.datetimes(), available_freqs(), st.text())
    def test_timedelta_index_as_input(
        self,
        timedelta_index_series: pd.Series,
        start: pd.datetime,
        freq: pd.Timedelta,
        output_name: str,
    ):
        time_series_preparation = TimeSeriesPreparation(
            start=start, freq=freq, output_name=output_name
        )
        computed_time_series = time_series_preparation.transform(timedelta_index_series)
        expected_series = timedelta_index_series_to_period_index_series(
            timedelta_index_series, freq=freq
        )
        expected_time_series = pd.DataFrame({output_name: expected_series})
        assert_frame_equal(computed_time_series, expected_time_series)



import pytest

import hypothesis.strategies as st
from hypothesis import given, settings, HealthCheck

from giottotime.feature_creation import ShiftFeature, MovingAverageFeature
from giottotime.model_selection.feature_splitters import FeatureSplitter
from giottotime.utils.hypothesis.feature_matrices import X_y_matrices


FEATURES = [
    ShiftFeature(0, "0"),
    ShiftFeature(1, "1"),
    MovingAverageFeature(3, "3"),
]

HORIZON = 4


class TestFeatureSplitter:
    def test_constructor(self):
        FeatureSplitter()

    @given(st.text().filter(lambda x: x != "any"))
    def test_constructor_wrong_parameter(self, drop_na_mode: str):
        with pytest.raises(ValueError):
            FeatureSplitter(drop_na_mode)

    @settings(suppress_health_check=(HealthCheck.too_slow,))
    @given(
        X_y_matrices(
            horizon=HORIZON, time_series_features=FEATURES, allow_nan_infinity=False,
        )
    )
    def test_transform(self, X_y):
        X, y = X_y
        feature_splitter = FeatureSplitter()
        X_train, y_train, X_test, y_test = feature_splitter.transform(X, y)

        assert X_train.shape[0] == max(0, X.shape[0] - 2 - HORIZON)
        assert y_train.shape[0] == X_train.shape[0]
        assert X_test.shape[0] == min(max(0, X.shape[0] - 2), HORIZON)
        assert y_test.shape[0] == X_test.shape[0]



from abc import ABCMeta, abstractmethod
from collections import Iterable
from typing import Union, List, Optional

import numpy as np
from giotto.time_series import TakensEmbedding, SlidingWindow
import giotto.diagrams as diag
import giotto.homology as hl
import pandas as pd

from ..base import IndexDependentFeature

__all__ = ["TDAFeatures"]


def _align_indices(
    X: pd.DataFrame, n_points: int, tda_feature_values: Union[List, np.ndarray]
) -> pd.DataFrame:
    output_X = X.copy()

    output_X.iloc[:-n_points] = np.nan

    splits = np.array_split(
        output_X.iloc[-n_points:].index.values, len(tda_feature_values)
    )

    for index, split in enumerate(splits):
        if isinstance(tda_feature_values[index], list) or isinstance(
            tda_feature_values[index], np.ndarray
        ):
            target_value = tda_feature_values[index][0]
        else:
            target_value = tda_feature_values[index]
        output_X.loc[split] = target_value

    return output_X


class TDAFeatures(IndexDependentFeature, metaclass=ABCMeta):
    """Base class for all the TDA feature_creation contained in the package.

    """

    @abstractmethod
    def __init__(
        self,
        output_name: str,
        takens_parameters_type: str = "search",
        takens_dimension: int = 5,
        takens_stride: int = 1,
        takens_time_delay: int = 1,
        takens_n_jobs: int = 1,
        sliding_window_width: int = 10,
        sliding_stride: int = 1,
        diags_metric: str = "euclidean",
        diags_coeff: int = 2,
        diags_max_edge_length: float = np.inf,
        diags_homology_dimensions: Iterable = (0, 1, 2),
        diags_infinity_values: float = None,
        diags_n_jobs: int = 1,
    ):
        super().__init__(output_name)

        self._takens_embedding = TakensEmbedding(
            parameters_type=takens_parameters_type,
            dimension=takens_dimension,
            stride=takens_stride,
            time_delay=takens_time_delay,
            n_jobs=takens_n_jobs,
        )
        self.takens_dimension = takens_dimension
        self.takens_stride = takens_stride
        self.takens_time_delay = takens_time_delay
        self.takens_dimension = takens_dimension

        self.sliding_window = SlidingWindow(
            width=sliding_window_width, stride=sliding_stride
        )
        self.sliding_window_width = sliding_window_width
        self.sliding_stride = sliding_stride

        self.vietoris_rips_persistence = hl.VietorisRipsPersistence(
            metric=diags_metric,
            coeff=diags_coeff,
            max_edge_length=diags_max_edge_length,
            homology_dimensions=diags_homology_dimensions,
            infinity_values=diags_infinity_values,
            n_jobs=diags_n_jobs,
        )

    def fit(self, time_series: pd.DataFrame, y: Optional[pd.DataFrame] = None):
        return self

    def _compute_n_points(self, n_windows: int) -> int:
        if n_windows <= 0:
            raise ValueError(
                f"The number of windows should be greater than "
                f"0, instead was {n_windows}."
            )
        embedder_length = (
            self.sliding_stride * (n_windows - 1) + self.sliding_window_width
        )

        n_used_points = (
            self.takens_stride * (embedder_length - 1)
            + self.takens_dimension * self.takens_time_delay
        )

        return n_used_points

    def _compute_persistence_diagrams(self, X: pd.DataFrame) -> np.ndarray:
        X_embedded = self._takens_embedding.fit_transform(X)
        self.X_embedded_dims_ = X_embedded.shape

        X_windows = self.sliding_window.fit_transform(X_embedded)
        X_diagrams = self.vietoris_rips_persistence.fit_transform(X_windows)

        diagram_scaler = diag.Scaler()
        diagram_scaler.fit(X_diagrams)

        return diagram_scaler.transform(X_diagrams)



from itertools import product

import pandas as pd
import numpy as np
from sklearn.utils.validation import check_is_fitted

from giottotime.causality_tests.base import CausalityTest


class ShiftedPearsonCorrelation(CausalityTest):
    """Class responsible for assessing the shifted Pearson correlations (PPMCC) between
    two or more series.

    Parameters
    ----------
    max_shift : int, optional, default: ``10``

    target_col : str, optional, default: ``'y'``
            The column to use as the a reference (i.e., the columns which is not
            shifted).

    dropna : bool, optional, default: ``False``
        Determines if the Nan values created by shifting are retained or dropped.

    Examples
    --------
    >>> from giottotime.causality_tests.shifted_pearson_correlation import ShiftedPearsonCorrelation
    >>> import pandas.util.testing as testing
    >>> data = testing.makeTimeDataFrame(freq="s")
    >>> spc = ShiftedPearsonCorrelation(target_col="A")
    >>> spc.fit(data)
    >>> spc.best_shifts_
    y  A  B  C  D
    x
    A  8  9  6  5
    B  7  4  4  6
    C  3  4  9  9
    D  7  1  9  1
    >>> spc.max_corrs_
    y         A         B         C         D
    x
    A  0.383800  0.260627  0.343628  0.360151
    B  0.311608  0.307203  0.255969  0.298523
    C  0.373613  0.267335  0.211913  0.140034
    D  0.496535  0.204770  0.402473  0.310065
    """

    def __init__(
        self,
        max_shift: int = 10,
        target_col: str = "y",
        dropna: bool = False,
        bootstrap_iterations=1000,
        bootstrap_samples=100,
    ):
        super().__init__(bootstrap_iterations, bootstrap_samples)
        self.max_shift = max_shift
        self.target_col = target_col
        self.dropna = dropna

    def fit(self, data: pd.DataFrame) -> "ShiftedPearsonCorrelation":
        """Create the dataframe of shifts of each time series which maximize the
         Pearson correlation (PPMCC).

        Parameters
        ----------
        data : pd.DataFrame, shape (n_samples, n_time_series), required
            The DataFrame containing the time series on which to compute the shifted
            correlations.

        Returns
        -------
        self : ``ShiftedPearsonCorrelation``

        """
        best_shifts = pd.DataFrame(columns=["x", "y", "shift", "max_corr", "p_values"])
        best_shifts = best_shifts.astype(
            {
                "x": np.float64,
                "y": np.float64,
                "shift": np.int64,
                "max_corr": np.int64,
                "p_values": np.float64,
            }
        )

        for x, y in product(data.columns, repeat=2):
            res = self._get_max_corr_shift(data, self.max_shift, x=x, y=y)
            best_shift = res[1]
            max_corr = res[0]
            p_value = self._compute_is_test_significant(data, x, y, best_shift)
            # N = data.shape[0] - max_shift

            best_shifts = best_shifts.append(
                {
                    "x": x,
                    "y": y,
                    "shift": best_shift,
                    "max_corr": max_corr,
                    "p_values": p_value,
                },
                ignore_index=True,
            )

        pivot_best_shifts = pd.pivot_table(
            best_shifts, index=["x"], columns=["y"], values="shift"
        )
        max_corrs = pd.pivot_table(
            best_shifts, index=["x"], columns=["y"], values="max_corr"
        )
        p_values = pd.pivot_table(
            best_shifts, index=["x"], columns=["y"], values="p_values"
        )

        self.best_shifts_ = pivot_best_shifts
        self.max_corrs_ = max_corrs
        self.p_values_ = p_values

        return self

    def transform(self, data: pd.DataFrame) -> pd.DataFrame:
        """Shifts each input time series by the amount which optimizes correlation with
        the selected 'y' column.

        Parameters
        ----------
        data : pd.DataFrame, shape (n_samples, n_time_series), required
            The DataFrame containing the time series on which to perform the
            transformation.

        Returns
        -------
        data_t : pd.DataFrame, shape (n_samples, n_time_series)
            The DataFrame (Pivot table) of the shifts which maximize the correlation
            between each time series The shift is indicated in rows.

        """
        check_is_fitted(self)
        data_t = data.copy()

        for col in data_t:
            if col != self.target_col:
                data_t[col] = data_t[col].shift(
                    -self.best_shifts_[col][self.target_col]
                )

        if self.dropna:
            data_t = data_t.dropna()

        return data_t

    def _get_max_corr_shift(
        self, data: pd.DataFrame, max_shift: int, x: str = "x", y: str = "y"
    ):
        shifts = pd.DataFrame()

        for shift in range(1, max_shift):
            shifts[shift] = data[x].shift(shift)

        shifts = shifts.dropna()
        self.shifted_corrs = shifts.corrwith(data[y])

        q = self.shifted_corrs.max(), self.shifted_corrs.idxmax()
        return q



import inspect
from abc import ABCMeta, abstractmethod


class TrendModel(metaclass=ABCMeta):
    @abstractmethod
    def fit(self, time_series):
        raise NotImplementedError  # to exclude from pytest coverage

    @abstractmethod
    def predict(self, t):
        raise NotImplementedError  # to exclude from pytest coverage

    @abstractmethod
    def transform(self, time_series):
        raise NotImplementedError  # to exclude from pytest coverage

    def __repr__(self):
        constructor_attributes = inspect.getfullargspec(self.__init__).args
        attributes_to_print = [
            str(getattr(self, attribute))
            for attribute in constructor_attributes
            if attribute in self.__dict__
        ]
        return f"{self.__class__.__name__}({', '.join(attributes_to_print)})"



from abc import abstractmethod
from typing import Union

import pandas as pd


class Feature:
    """Base class for all the features defined in this package.

    """

    @abstractmethod
    def __init__(self, output_name):
        self.output_name = output_name

    def _rename_columns(self, X: Union[pd.DataFrame, pd.Series]) -> pd.DataFrame:
        if isinstance(X, pd.Series):
            X = X.to_frame()

        suffix = ""

        X_renamed = X.T.reset_index(drop=True).T
        for index, col in enumerate(X_renamed.columns):
            if len(X.columns) > 1:
                suffix = "_" + str(index)
            X_renamed.rename(
                columns={col: self.output_name + str(suffix)}, inplace=True
            )

        return X_renamed



from typing import Optional, Dict, Union, Iterable

import hypothesis.strategies as st
import numpy as np
import pandas as pd
from hypothesis import assume
from hypothesis.extra.numpy import arrays

from .utils import (
    freq_to_timedelta,
    initialize_start_date_end_date,
    order_pair,
    initialize_start_timedelta_end_timedelta,
    expected_index_length_from,
    expected_start_date_from,
    expected_end_date_from,
)

IndexRangeArgs = Dict[str, Union[pd.datetime, int, pd.Timedelta]]

string_freqs = ["B", "D", "W", "M", "Q", "A", "Y", "H", "T", "S"]
pandas_range_params = ["start", "end", "periods", "freq"]


@st.composite
def giotto_time_series(
    draw,
    start_date: Optional[pd.Timestamp] = None,
    end_date: Optional[pd.Timestamp] = None,
    min_length: int = 0,
    max_length: int = 1000,
    name: str = "time_series",
    allow_nan: bool = True,
    allow_infinity: bool = True,
):
    period_index_series = draw(
        series_with_period_index(
            start_date,
            end_date,
            min_length,
            max_length,
            allow_nan=allow_nan,
            allow_infinity=allow_infinity,
        )
    )
    return pd.DataFrame({name: period_index_series})


@st.composite
def series_with_period_index(
    draw,
    start: Optional[pd.datetime] = None,
    end: Optional[pd.datetime] = None,
    min_length: int = 0,
    max_length: int = 1000,
    allow_nan: bool = True,
    allow_infinity: bool = True,
):
    """ Returns a strategy to generate a Pandas Series with PeriodIndex

    Parameters
    ----------
    draw
    start : ``pd.datetime``, optional, (default=None)
    end : ``pd.datetime``, optional, (default=None)
    min_length : ``int``, optional, (default=0)
    max_length : ``int``, optional, (default=None)
    allow_nan : ``bool``, optional, (default=True)
    allow_infinity : ``bool``, optional, (default=True)

    Returns
    -------
    pd.Series with PeriodIndex
    """
    index = draw(period_indexes(start, end, min_length, max_length))
    values = draw(
        arrays(
            dtype=np.float64,
            shape=index.shape[0],
            elements=st.floats(allow_nan=allow_nan, allow_infinity=allow_infinity),
        )
    )
    return pd.Series(index=index, data=values)


@st.composite
def series_with_datetime_index(
    draw,
    start: Optional[pd.datetime] = None,
    end: Optional[pd.datetime] = None,
    min_length: int = 0,
    max_length: int = 1000,
    allow_nan: bool = True,
    allow_infinity: bool = True,
):
    """ Returns a strategy to generate a Pandas Series with DatetimeIndex

    Parameters
    ----------
    draw
    start : ``pd.datetime``, optional, (default=None)
    end : ``pd.datetime``, optional, (default=None)
    min_length : ``int``, optional, (default=0)
    max_length : ``int``, optional, (default=1000)
    allow_nan : ``bool``, optional, (default=True)
    allow_infinity : ``bool``, optional, (default=True)

    Returns
    -------
    pd.Series with DatetimeIndex
    """
    index = draw(datetime_indexes(start, end, min_length, max_length))
    values = draw(
        arrays(
            dtype=np.float64,
            shape=index.shape[0],
            elements=st.floats(allow_nan=allow_nan, allow_infinity=allow_infinity),
        )
    )
    return pd.Series(index=index, data=values)


@st.composite
def series_with_timedelta_index(
    draw,
    start: Optional[pd.Timedelta] = None,
    end: Optional[pd.Timedelta] = None,
    min_length: int = 0,
    max_length: int = 1000,
    allow_nan: bool = True,
    allow_infinity: bool = True,
):
    """ Returns a strategy to generate a Pandas Series with Timedelta index

    Parameters
    ----------
    draw
    start : ``pd.Timedelta``, optional, (default=None)
    end : ``pd.Timedelta``, optional, (default=None)
    min_length : ``int``, optional, (default=0)
    max_length : ``int``, optional, (default=None)
    allow_nan : ``bool``, optional, (default=True)
    allow_infinity : ``bool``, optional, (default=True)

    Returns
    -------
    pd.Series with TimedeltaIndex
    """
    index = draw(timedelta_indexes(start, end, min_length, max_length))
    values = draw(
        arrays(
            dtype=np.float64,
            shape=index.shape[0],
            elements=st.floats(allow_nan=allow_nan, allow_infinity=allow_infinity),
        )
    )
    return pd.Series(index=index, data=values)


@st.composite
def period_indexes(
    draw,
    start: Optional[pd.datetime] = None,
    end: Optional[pd.datetime] = None,
    min_length: int = 0,
    max_length: int = 1000,
):
    """ Returns a strategy to generate Pandas PeriodIndex

    Parameters
    ----------
    draw
    start : ``pd.datetime``, optional, (default=None)
    end : ``pd.datetime``, optional, (default=None)
    min_length : ``int``, optional, (default=0)
    max_length : ``int``, optional, (default=1000)

    Returns
    -------
    LazyStrategy that generates pandas PeriodIndex
    """
    period_range_args = draw(period_range_args_from(start, end, min_length, max_length))
    assume(
        period_range_args_are_correct(
            period_range_args, min_length, max_length, start, end
        )
    )
    return compute_period_range(period_range_args)


@st.composite
def datetime_indexes(
    draw,
    start: Optional[pd.datetime] = None,
    end: Optional[pd.datetime] = None,
    min_length: int = 0,
    max_length: int = 1000,
):
    """ Returns a strategy to generate Pandas DatetimeIndex.

    Parameters
    ----------
    draw
    start : ``pd.datetime``, optional, (default=None)
    end : ``pd.datetime``, optional, (default=None)
    min_length : ``int``, optional, (default=0)
    max_length : ``int``, default: ``1000``

    Returns
    -------
    LazyStrategy that generates pd.DatetimeIndex
    """
    date_range_args = draw(date_range_args_from(start, end, min_length, max_length))
    assume(
        date_range_args_are_correct(date_range_args, min_length, max_length, start, end)
    )
    return compute_date_range(date_range_args)


@st.composite
def timedelta_indexes(
    draw,
    start: Optional[pd.Timedelta] = None,
    end: Optional[pd.Timedelta] = None,
    min_length: int = 0,
    max_length: int = 1000,
):
    """ Returns a strategy to generate Pandas TimedeltaIndex

    Parameters
    ----------
    draw
    start : ``pd.Timedelta``, optional, (default=None)
    end : ``pd.Timedelta``, optional, (default=None)
    min_length : ``int``, optional, (default=0)
    max_length : ``int``, (default=1000)

    Returns
    -------
    LazyStrategy that generates pd.TimedeltaIndex
    """
    timedelta_range_args = draw(
        timedelta_range_args_from(start, end, min_length, max_length)
    )
    assume(
        timedelta_range_args_are_correct(
            timedelta_range_args,
            min_length=min_length,
            max_length=max_length,
            min_start=start,
            max_end=end,
        )
    )
    return compute_timedelta_range(timedelta_range_args)


def pair_of_ordered_dates(
    start: Optional[pd.datetime] = None, end: Optional[pd.datetime] = None
):
    """ Returns an hypothesis strategy that generates a pair of ordered
    pd.datetime. Useful to create a Pandas index

    Parameters
    ----------
    start : ``pd.datetime``, optional, (default=None)
    end : ``pd.datetime``, optional, (default=None)

    Returns
    -------
    LazyStrategy that generates Tuple[pd.datetime, pd.datetime]
    """
    start, end = initialize_start_date_end_date(start, end)

    date1 = st.datetimes(start, end)
    date2 = st.datetimes(start, end)

    return order_pair(date1, date2)


def pair_of_ordered_timedeltas(
    start: Optional[pd.Timedelta] = None, end: Optional[pd.Timedelta] = None
):
    """ Returns an hypothesis strategy that generates a pair of ordered
    pd.Timedelta. Useful to create a Pandas TimedeltaIndex

    Parameters
    ----------
    start : ``pd.datetime``, optional, (default=None)
    end : ``pd.datetime``, optional, (default=None)

    Returns
    -------
    LazyStrategy that generates Tuple[pd.Timedelta, pd.Timedelta]
    """
    start, end = initialize_start_timedelta_end_timedelta(start, end)

    timedelta1 = st.timedeltas(start, end)
    timedelta2 = st.timedeltas(start, end)

    return order_pair(timedelta1, timedelta2)


def positive_bounded_integers(max_value):
    """ Returns a strategy that generates a positive bounded integer

    Parameters
    ----------
    max_value : ``int``, required
        maximum value of the integer

    Returns
    -------
    LazyStrategy
    """
    return st.integers(min_value=0, max_value=max_value)


def samples_from(iterable: Iterable):
    """ Strategy that samples from an iterable.

    Parameters
    ----------
    iterable : ``Iterable[Generic]``, required

    Returns
    -------
    LazyStrategy
    """
    return st.builds(lambda index: iterable[index], st.integers(0, len(iterable) - 1))


@st.composite
def available_freqs(draw) -> pd.Timedelta:
    """ Strategy that samples from the available frequencies and returns a
    corresponding pd.Timedelta

    Parameters
    ----------
    draw

    Returns
    -------
    pd.Timedelta
    """
    return freq_to_timedelta(draw(samples_from(string_freqs)))


@st.composite
def period_range_args_from(
    draw,
    start: pd.datetime,
    end: pd.datetime,
    min_length: int,
    max_length: int,
    arg_to_remove: Optional[str] = None,
) -> IndexRangeArgs:
    """ Strategy that computes arguments to call pd.period_range() from
    start, end and max length of the resulting period
    It returns a dictionary with 3 out of the 3 keys 'start', 'end', 'periods'
    and 'freq'. One of them is automatically removed since pd.period_range()
    accepts only 3 of the 4 inputs.

    Parameters
    ----------
    draw
    start : ``pd.datetime``, required
    end : ``pd.datetime``, required
    min_length : ``int``, required
    max_length : ``int``, required
    arg_to_remove : ``str``, optional, (default=None)

    Returns
    -------
    IndexRangeArgs
        dictionary with 3 out of the 4 keys 'start', 'end', 'periods' and
        'freq'.
    """
    start, end = draw(pair_of_ordered_dates(start, end))
    periods = draw(st.integers(min_value=min_length, max_value=max_length))
    freq = draw(available_freqs())
    element_to_exclude = (
        draw(samples_from(pandas_range_params))
        if arg_to_remove is None
        else arg_to_remove
    )
    assume(element_to_exclude != "freq")

    args = {"start": start, "end": end, "periods": periods, "freq": freq}
    del args[element_to_exclude]

    return args


def period_range_args_are_correct(
    period_range_args: IndexRangeArgs,
    min_length: int = 0,
    max_length: int = 1000,
    min_start: Optional[pd.Period] = None,
    max_end: Optional[pd.Period] = None,
) -> bool:
    """ Returns True if the period range arguments are correct. i.e. if
    'periods' is not a key it checks that the length of the resulting period
    index is below ``max_length``.

    Parameters
    ----------
    period_range_args : ``IndexRangeArgs``, required
        dictionary with 3 out of the 4 keys 'start', 'end', 'periods' and
        'freq'.

    min_length : ``int``, optional, (default=0)
    max_length : ``int``, optional, (default=1000)
    min_start : ``pd.Period``, optional, (default=1000)
    max_end : ``pd.Period``, optional, (default=1000


    Returns
    -------
    bool
    """
    min_start = min_start if min_start is not None else pd.Timestamp("1980-01-01")
    max_end = max_end if max_end is not None else pd.Timestamp("2020-01-01")
    return _range_args_are_correct(
        period_range_args, min_length, max_length, min_start, max_end
    )


def compute_period_range(period_range_args: IndexRangeArgs) -> pd.PeriodIndex:
    """ Returns a Pandas PeriodIndex from the given ``period_range_args``.

    Parameters
    ----------
    period_range_args : ``IndexRangeArgs``, required

    Returns
    -------
    pd.PeriodIndex
    """
    return pd.period_range(**period_range_args)


@st.composite
def date_range_args_from(
    draw,
    start: pd.datetime,
    end: pd.datetime,
    min_length: int,
    max_length: int,
    arg_to_remove: Optional[str] = None,
) -> IndexRangeArgs:
    """ Strategy that computes arguments to call pd.date_range() from
    start, end and max length of the resulting period
    It returns a dictionary with 3 out of the 3 keys 'start', 'end', 'periods'
    and 'freq'. One of them is automatically removed since pd.date_range()
    accepts only 3 of the 4 inputs.

    Parameters
    ----------
    draw
    start : ``pd.datetime``, required
    end : ``pd.datetime``, required
    min_length: ``int``, required
    max_length : ``int``, required
    arg_to_remove: ``str``, optional, (default=None)

    Returns
    -------
    IndexRangeArgs
        dictionary with 3 out of the 4 keys 'start', 'end', 'periods' and
        'freq'.
    """
    start, end = draw(pair_of_ordered_dates(start, end))
    periods = draw(st.integers(min_value=min_length, max_value=max_length))
    freq = draw(available_freqs())
    element_to_exclude = (
        draw(samples_from(pandas_range_params))
        if arg_to_remove is None
        else arg_to_remove
    )

    args = {"start": start, "end": end, "periods": periods, "freq": freq}
    del args[element_to_exclude]

    return args


def date_range_args_are_correct(
    date_range_args: IndexRangeArgs,
    min_length: int = 0,
    max_length: int = 1000,
    min_start: pd.Timestamp = None,
    max_end: pd.Timestamp = None,
) -> bool:
    """ Returns True if the date range arguments are correct.
    It checks 3 things:

    - if 'periods' is not a key it checks that the length of the resulting
    period index is below ``max_length``.
    - if 'start' is not a keyword it checks that the expected start date is
    after ``min_start``
    - if 'end' is not a keyword it checks that the expected end date is
    before ``max_end``

    Parameters
    ----------
    date_range_args : ``IndexRangeArgs``, required
        dictionary with 3 out of the 4 keys 'start', 'end', 'periods' and
        'freq'.

    min_length : ``int``, optional, (default=0)
    max_length : ``int``, optional, (default=1000)
    min_start : ``pd.Timestamp``, optional, (default=None)
    max_end : ``pd.Timestamp``, optional, (default=None)

    Returns
    -------
    bool
    """
    min_start = min_start if min_start is not None else pd.Timestamp("1980-01-01")
    max_end = max_end if max_end is not None else pd.Timestamp("2020-01-01")
    return _range_args_are_correct(
        date_range_args, min_length, max_length, min_start, max_end
    )


def compute_date_range(date_range_args: IndexRangeArgs) -> pd.DatetimeIndex:
    """ Returns a Pandas DatetimeIndex from the given ``date_range_args``.

    Parameters
    ----------
    date_range_args: ``IndexRangeArgs``, required

    Returns
    -------
    pd.DatetimeIndex
    """
    return pd.date_range(**date_range_args)


@st.composite
def timedelta_range_args_from(
    draw,
    start: pd.Timedelta,
    end: pd.Timedelta,
    min_length: int,
    max_length: int,
    arg_to_remove: Optional[str] = None,
) -> IndexRangeArgs:
    """ Strategy that computes arguments to call pd.timedelta_range() from
    start, end and max length of the resulting period
    It returns a dictionary with 3 out of the 3 keys 'start', 'end', 'periods'
    and 'freq'. One of them is automatically removed since pd.timedelta_range()
    accepts only 3 of the 4 inputs.

    Parameters
    ----------
    draw
    start : pd.datetime, required
    end : pd.datetime, required
    min_length : int, required
    max_length : int, required
    arg_to_remove: str, optional, (default=None)

    Returns
    -------
    IndexRangeArgs
        dictionary with 3 out of the 4 keys 'start', 'end', 'periods' and
        'freq'.
    """
    start, end = draw(pair_of_ordered_timedeltas(start, end))
    periods = draw(st.integers(min_value=min_length, max_value=max_length))
    freq = draw(available_freqs())
    element_to_exclude = (
        draw(samples_from(pandas_range_params))
        if arg_to_remove is None
        else arg_to_remove
    )

    args = {"start": start, "end": end, "periods": periods, "freq": freq}
    del args[element_to_exclude]

    return args


def timedelta_range_args_are_correct(
    timedelta_range_args: IndexRangeArgs,
    min_length: int = 0,
    max_length: int = 1000,
    min_start: pd.Timedelta = None,
    max_end: pd.Timedelta = None,
) -> bool:
    """ Returns True if the timedelta range arguments are correct.
    It checks 3 things:

    - if 'periods' is not a key it checks that the length of the resulting
    period index is below ``max_length``.
    - if 'start' is not a keyword it checks that the expected start timedelta
    is after ``min_start``
    - if 'end' is not a keyword it checks that the expected end timedelta is
    before ``max_end``

    Parameters
    ----------
    timedelta_range_args : ``IndexRangeArgs``, required
        dictionary with 3 out of the 4 keys 'start', 'end', 'periods' and
        'freq'.

    min_length : ``int``, optional, (default=0)
    max_length : ``int``, optional, (default=1000)
    min_start : ``pd.Timestamp``, optional, (default=None)
    max_end : ``pd.Timestamp``, optional, (default=None)

    Returns
    -------
    bool
    """
    min_start = min_start if min_start is not None else pd.Timedelta(0)
    max_end = max_end if max_end is not None else pd.Timedelta("40Y")
    return _range_args_are_correct(
        timedelta_range_args, min_length, max_length, min_start, max_end
    )


def compute_timedelta_range(timedelta_range_args: IndexRangeArgs):
    """ Returns a Pandas TimedeltaIndex from the given 
    ``timedelta_range_args``.

    Parameters
    ----------
    timedelta_range_args: ``IndexRangeArgs``, required

    Returns
    -------
    pd.DatetimeIndex
    """
    return pd.timedelta_range(**timedelta_range_args)


def _range_args_are_correct(
    range_args: IndexRangeArgs,
    min_length: int = 0,
    max_length: int = 1000,
    min_start: Union[pd.Timedelta, pd.datetime] = None,
    max_end: Union[pd.Timedelta, pd.datetime] = None,
) -> bool:
    """Checks if range args are correct.

    Parameters
    ----------
    range_args: ``IndexRangeArgs``, required
    min_length : ``int``, optional, (default=0)
    max_length : ``int``, optional, (default=1000)
    min_start : ``Union[pd.Timedelta, pd.datetime]``, optional, (default=None)
    max_end : ``Union[pd.Timedelta, pd.datetime]``, optional, (default=None)

    Returns
    -------
    bool
    """
    try:
        if "periods" not in range_args:
            return min_length < expected_index_length_from(**range_args) < max_length
        elif "start" not in range_args:
            return expected_start_date_from(**range_args) >= min_start
        elif "end" not in range_args:
            return expected_end_date_from(**range_args) <= max_end
        else:
            return True
    except OverflowError:
        return False



"""
Time series module for Python
==================================
giottotime is a Python module integrating time series and machine learning.

See URL TO GIOTTOTIME for complete documentation.
"""

from giottotime._version import __version__


__all__ = [
    "causality_tests",
    "feature_creation",
    "loss_functions",
    "model_selection",
    "models",
    "time_series_preparation",
]



from abc import ABCMeta, abstractmethod

import pandas as pd
from scipy import stats


class CausalityTest(metaclass=ABCMeta):
    """ Base class for causality tests.

    """

    def __init__(self, bootstrap_iterations, bootstrap_samples):
        self.bootstrap_iterations = bootstrap_iterations
        self.bootstrap_samples = bootstrap_samples

    @abstractmethod
    def fit(self, data_matrix):
        raise NotImplementedError  # to exclude from pytest coverage

    @abstractmethod
    def transform(self, time_series):
        raise NotImplementedError  # to exclude from pytest coverage

    def _compute_is_test_significant(self, data, x, y, best_shift):
        bootstrap_matrix = data.copy()
        bootstrap_matrix[y] = bootstrap_matrix.shift(best_shift)[y]
        bootstrap_matrix.dropna(axis=0, inplace=True)
        rhos = []

        for k in range(self.bootstrap_iterations):
            bootstraps = bootstrap_matrix.sample(n=len(data), replace=True)
            rhos.append(stats.pearsonr(bootstraps[x], bootstraps[y])[0])
        rhos = pd.DataFrame(rhos)

        percentile = stats.percentileofscore(rhos, 0) / 100
        p_value = 2 * (percentile if percentile < 0.5 else 1 - percentile)

        return p_value



from typing import Iterable

import giotto.diagrams as diag
import numpy as np
import pandas as pd
import pytest
from hypothesis import given, strategies as st

from giottotime.feature_creation.index_dependent_features.tda_features import (
    TDAFeatures,
)


class BaseTDAFeature(TDAFeatures):
    def __init__(
        self,
        output_name: str,
        takens_parameters_type: str = "search",
        takens_dimension: int = 5,
        takens_stride: int = 1,
        takens_time_delay: int = 1,
        takens_n_jobs: int = 1,
        sliding_window_width: int = 10,
        sliding_stride: int = 1,
        diags_metric: str = "euclidean",
        diags_coeff: int = 2,
        diags_max_edge_length: float = np.inf,
        diags_homology_dimensions: Iterable = (0, 1, 2),
        diags_infinity_values: float = None,
        diags_n_jobs: int = 1,
    ):
        super().__init__(
            output_name,
            takens_parameters_type,
            takens_dimension,
            takens_stride,
            takens_time_delay,
            takens_n_jobs,
            sliding_window_width,
            sliding_stride,
            diags_metric,
            diags_coeff,
            diags_max_edge_length,
            diags_homology_dimensions,
            diags_infinity_values,
            diags_n_jobs,
        )

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        pass

    def _correct_n_points_formula(self, n_windows: int) -> int:
        embedder_length = (
            self.sliding_stride * (n_windows - 1) + self.sliding_window_width
        )

        n_used_points = self.takens_stride * (embedder_length - 1) + (
            self.takens_dimension * self.takens_time_delay
        )
        return n_used_points

    def _correct_compute_persistence_diagrams(self, X: pd.DataFrame) -> np.ndarray:
        X_embedded = self._takens_embedding.fit_transform(X,)
        self.X_embedded_dims_ = X_embedded.shape

        X_windows = self.sliding_window.fit_transform(X_embedded,)
        X_diagrams = self.vietoris_rips_persistence.fit_transform(X_windows,)

        diagram_scaler = diag.Scaler()
        diagram_scaler.fit(X_diagrams)

        return diagram_scaler.transform(X_diagrams)


def test_correct_compute_n_points():
    sliding_stride = 3
    sliding_window_width = 2
    takens_stride = 5
    takens_dimension = 2
    takens_time_delay = 3
    n_windows = 9

    tda_feature = BaseTDAFeature(
        output_name="ignored",
        sliding_stride=sliding_stride,
        sliding_window_width=sliding_window_width,
        takens_stride=takens_stride,
        takens_dimension=takens_dimension,
        takens_time_delay=takens_time_delay,
    )
    n_points = tda_feature._compute_n_points(n_windows)
    expected_n_points = 131

    assert expected_n_points == n_points


def test_negative_or_zero_n_windows():
    tda_feature = BaseTDAFeature(output_name="ignored")

    with pytest.raises(ValueError):
        tda_feature._compute_n_points(0)

    with pytest.raises(ValueError):
        tda_feature._compute_n_points(-4)


@given(
    st.integers(1, 20),
    st.integers(1, 20),
    st.integers(1, 20),
    st.integers(1, 20),
    st.integers(1, 20),
    st.integers(1, 20),
)
def test_correct_n_points_random_ts_and_values(
    n_windows,
    sliding_stride,
    sliding_window_width,
    takens_stride,
    takens_dimension,
    takens_time_delay,
):
    tda_feature = BaseTDAFeature(
        output_name="ignored",
        sliding_stride=sliding_stride,
        sliding_window_width=sliding_window_width,
        takens_stride=takens_stride,
        takens_dimension=takens_dimension,
        takens_time_delay=takens_time_delay,
    )
    n_points = tda_feature._compute_n_points(n_windows)
    expected_n_points = tda_feature._correct_n_points_formula(n_windows)
    assert expected_n_points == n_points
    assert expected_n_points > 0


def test_correct_persistence_diagrams():
    np.random.seed(0)
    df = pd.DataFrame(np.random.randint(0, 100, size=(13, 1)), columns=list("A"))

    tda_feature = BaseTDAFeature(output_name="ignored")
    persistence_diagrams = tda_feature._compute_persistence_diagrams(df)
    expected_diagrams = np.array(
        [
            [
                [0.0, 0.435255, 0.0],
                [0.0, 0.44177876, 0.0],
                [0.0, 0.62426053, 0.0],
                [0.0, 0.64725938, 0.0],
                [0.0, 0.76393602, 0.0],
                [0.0, 0.86575012, 0.0],
                [0.0, 1.21810876, 0.0],
                [0.0, 1.28787198, 0.0],
                [0.0, 1.40406276, 0.0],
                [0.0, 1.41421356, 0.0],
                [1.54689366, 1.57882526, 1.0],
                [1.46434634, 1.4783885, 1.0],
                [0.0, 0.0, 2.0],
            ]
        ]
    )

    assert np.allclose(expected_diagrams, persistence_diagrams)



from typing import Callable

from sklearn.metrics import mean_squared_error
from scipy.optimize import minimize

import numpy as np
import pandas as pd
from sklearn.utils.validation import check_is_fitted

from .base import TrendModel


class ExponentialTrend(TrendModel):
    """A model for fitting, predicting and removing an exponential trend from a time
    series.

    Parameters
    ----------
    loss : Callable, optional, default: ``mean_squared_error``
        The loss function to use when fitting the model. The loss function must accept
        y_true, y_pred and return a single real number.

    method : str, optional, default: ``'BFGS``
            The method to use in order to minimize the loss function.

    """

    def __init__(self, loss: Callable = mean_squared_error, method: str = "BFGS"):
        self.loss = loss
        self.method = method

    def fit(self, time_series: pd.DataFrame) -> TrendModel:
        """Fit the model on the ``time_series``, with respect to the provided ``loss``
        and using the provided ``method``. In order to see which methods are available,
        please check the 'scipy' `documentation
        <https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.minimize.html>`_.

        Parameters
        ----------
        time_series: pd.DataFrame, shape (n_samples, 1), required
            The time series on which to fit the model.

        Returns
        -------
        self : TrendModel
            The fitted object.

        """

        def prediction_error(exponent):
            predictions = [np.exp(t * exponent) for t in range(0, time_series.shape[0])]
            return self.loss(time_series.values, predictions)

        model_exponent = 0
        res = minimize(
            prediction_error,
            np.array([model_exponent]),
            method=self.method,
            options={"disp": False},
        )

        self.model_exponent_ = res["x"][0]

        self.t0_ = time_series.index[0]
        freq = time_series.index.freq
        if freq is not None:
            self.period_ = freq
        else:
            self.period_ = time_series.index[1] - time_series.index[0]

        return self

    def predict(self, time_series: pd.DataFrame) -> pd.DataFrame:
        """Using the fitted model, predict the value starting from ``time_series``.

        Parameters
        ----------
        time_series: pd.DataFrame, shape (n_samples, 1), required
            The time series on which to predict.

        Returns
        -------
        predictions : pd.DataFrame, shape (n_samples, 1)
            The output predictions.

        Raises
        ------
        NotFittedError
            Raised if the model is not fitted yet.

        """
        check_is_fitted(self)

        predictions = np.exp(time_series * self.model_exponent_)
        return predictions

    def transform(self, time_series: pd.DataFrame) -> pd.DataFrame:
        """Transform the ``time_series`` by removing the trend.

        Parameters
        ----------
        time_series : ``pd.DataFrame``, required.
            The time series to transform.

        Returns
        -------
        transformed_time_series : ``pd.DataFrame``
            The transformed time series, without the trend.

        """
        ts = (time_series.index - self.t0_) / self.period_

        predictions = pd.Series(
            index=time_series.index,
            data=[np.exp(t * self.model_exponent_) for t in ts],
        )

        return time_series.sub(predictions, axis=0)



import random

import numpy as np
import pandas.util.testing as testing
import pytest
from hypothesis import given, strategies as st
from sklearn.exceptions import NotFittedError
from sklearn.linear_model import LinearRegression

from giottotime.feature_creation import (
    MovingAverageFeature,
    ConstantFeature,
    ShiftFeature,
)
from giottotime.feature_creation.feature_creation import FeatureCreation
from giottotime.model_selection import FeatureSplitter
from giottotime.models.time_series_models.gar import GAR


class TestInputs:
    @pytest.mark.parametrize(
        "base_model", [None, FeatureCreation(horizon=1, time_series_features=[])]
    )
    def test_invalid_base_model(self, base_model):
        with pytest.raises(TypeError):
            GAR(base_model=base_model, feed_forward=False)

        with pytest.raises(TypeError):
            GAR(base_model=base_model, feed_forward=True)


@pytest.fixture
def time_series():
    testing.N, testing.K = 200, 1
    return testing.makeTimeDataFrame(freq="MS")


def arbitrary_features(feature_length):
    possible_features = [MovingAverageFeature, ConstantFeature, ShiftFeature]
    random_features = []
    random_params = random.sample(range(1, 100), feature_length)

    feature_names = []
    for random_param in random_params:
        random_feature = random.sample(possible_features, 1)[0]
        output_name = f"{random_feature}_{random_params}"
        if output_name in feature_names:
            continue
        feature_names.append(output_name)
        feature = random_feature(random_param, output_name=output_name)
        random_features.append(feature)

    return random_features


class TestFitPredict:
    def test_predict_with_no_fit(self, time_series):
        base_model = LinearRegression()

        gar_no_feedforward = GAR(base_model=base_model, feed_forward=False)

        with pytest.raises(NotFittedError):
            gar_no_feedforward.predict(time_series)

        gar_feedforward = GAR(base_model=base_model, feed_forward=True)

        with pytest.raises(NotFittedError):
            gar_feedforward.predict(time_series)

    @given(st.builds(arbitrary_features, st.integers(1, 50)))
    def test_correct_features_dimension(self, time_series, features):
        horizon = 4
        feature_creation = FeatureCreation(
            horizon=horizon, time_series_features=features
        )
        base_model = LinearRegression()

        x, y = feature_creation.fit_transform(time_series)

        feature_splitter = FeatureSplitter()
        x_train, y_train, x_test, y_test = feature_splitter.transform(x, y)

        gar_no_feedforward = GAR(base_model=base_model, feed_forward=False)

        gar_no_feedforward.fit(x_train, y_train)
        assert gar_no_feedforward.train_features_.shape[1] == len(features)

        gar_feedforward = GAR(base_model=base_model, feed_forward=True)

        gar_feedforward.fit(x_train, y_train)
        assert gar_feedforward.train_features_.shape[1] == len(features)

    @given(st.builds(arbitrary_features, st.integers(1, 50)))
    def test_correct_fit_date(self, time_series, features):
        horizon = 4
        feature_creation = FeatureCreation(
            horizon=horizon, time_series_features=features
        )
        base_model = LinearRegression()

        x, y = feature_creation.fit_transform(time_series)

        feature_splitter = FeatureSplitter()
        x_train, y_train, x_test, y_test = feature_splitter.transform(x, y)

        gar_no_feedforward = GAR(base_model=base_model, feed_forward=False)

        gar_no_feedforward.fit(x_train, y_train)

        predictions = gar_no_feedforward.predict(x_test)

        assert len(predictions) == len(x_test)
        np.testing.assert_array_equal(predictions.index, x_test.index)

        gar_with_feedforward = GAR(base_model=base_model, feed_forward=True)

        gar_with_feedforward.fit(x_train, y_train)

        predictions = gar_with_feedforward.predict(x_test)

        assert len(predictions) == len(x_test)
        np.testing.assert_array_equal(predictions.index, x_test.index)



from typing import Iterable, List, Optional, Callable, Union

import pandas as pd
import numpy as np
import giotto.diagrams as diag

from .base import TDAFeatures, _align_indices


def _find_mean_nonzero(g):
    if g.to_numpy().nonzero()[1].any():
        return g.to_numpy().nonzero()[1].mean()
    else:
        return 0


class BettiCurvesFeature(TDAFeatures):
    """Compute the list of average lifetime for each time window, starting from the
    persistence diagrams.

    Parameters
    ----------
    betti_mode : ``'mean'`` | ``'arg_max'``, optional, default: ``'mean'``
        If ``mean``, compute the mean.

    output_name : str, optional, default: ``'BettiCurvesFeature'``
        The name of the output column.

    betti_homology_dimensions : ``Iterable`, optional, (default=``(0, 1)``)
        Dimensions (non-negative integers) of the topological feature_creation
        to be detected.

    betti_n_values : int, optional, default: ``100``
        The number of filtration parameter values, per available homology
        dimension, to sample during :meth:`fit`.

    betti_rolling : int, optional, default: ``1``
        Used only if ``betti_mode`` is set to ``mean``. When computing the
        betti surfaces, used to set the rolling parameter.

    betti_n_jobs : int, optional, default: ``None``
        The number of jobs to use for the computation. ``None`` means 1
        unless in a :obj:`joblib.parallel_backend` context. ``-1`` means
        using all processors.

    takens_parameters_type: ``'search'`` | ``'fixed'``, optional, default: ``'search'``
        If set to ``'fixed'``, the values of `time_delay` and `dimension are used
        directly in :meth:`transform`. If set to ``'search'``, those values are only
        used as upper bounds in a search as follows: first, an optimal time delay is
        found by minimising the time delayed mutual information; then, a heuristic based
        on an algorithm in [2]_ is used to select an embedding dimension which, when
        increased, does not reveal a large proportion of "false nearest neighbors".

    takens_time_delay : int, optional, default: ``1``
        Time delay between two consecutive values for constructing one embedded point.
        If `parameters_type` is ``'search'``, it corresponds to the maximal embedding
        time delay that will be considered.

    takens_dimension : int, optional, default: ``5``
        Dimension of the embedding space. If `parameters_type` is ``'search'``, it
        corresponds to the maximum embedding dimension that will be considered.

    takens_stride : int, optional, default: ``1``
        Stride duration between two consecutive embedded points. It defaults to 1 as
        this is the usual value in the statement of Takens's embedding theorem.

    takens_n_jobs : int: , optional, default: ``None``
        The number of jobs to use for the computation. ``None`` means 1 unless in a
        :obj:`joblib.parallel_backend` context. ``-1`` means using all processors.

    sliding_window_width : int, optional, default: ``10``
        Width of each sliding window. Each window contains ``width + 1`` objects from
        the original time series.

    sliding_stride : int, optional, default: ``1``
        Stride between consecutive windows.

    diags_metric : Union[str, Callable], optional, default: ``'euclidean'``
        If set to `'precomputed'`, input data is to be interpreted as a collection of
        distance matrices. Otherwise, input data is to be interpreted as a collection
        of point clouds (i.e. feature arrays), and `metric` determines a rule with which
        to calculate distances between pairs of instances (i.e. rows) in these arrays.
        If `metric` is a string, it must be one of the options allowed by
        :obj:`scipy.spatial.distance.pdist` for its metric parameter, or a metric listed
        in :obj:`sklearn.pairwise.PAIRWISE_DISTANCE_FUNCTIONS`, including "euclidean",
        "manhattan", or "cosine". If `metric` is a callable function, it is called on
        each pair of instances and the resulting value recorded. The callable should
        take two arrays from the entry in `X` as input, and return a value indicating
        the distance between them.

    diags_homology_dimensions : Iterable, optional, default: ``(0, 1)``
        Dimensions (non-negative integers) of the topological feature_creation to be
        detected.

    diags_coeff : int prime, optional, default: ``2``
        Compute homology with coefficients in the prime field
        :math:`\\mathbb{F}_p = \\{ 0, \\ldots, p - 1 \\}` where
        :math:`p` equals `coeff`.

    diags_max_edge_length : float, optional, default: ``np.inf``
        Upper bound on the maximum value of the Vietoris-Rips filtration parameter.
        Points whose distance is greater than this value will never be connected by an
        edge, and topological feature_creation at scales larger than this value will not
        be detected.

    diags_infinity_values : float, optional, default: ``None``
        Which death value to assign to feature_creation which are still alive at
        filtration value `max_edge_length`. ``None`` has the same behaviour as
        `max_edge_length`.

    diags_n_jobs : int, optional, default: ``None``
        The number of jobs to use for the computation. ``None`` means 1 unless in a
        :obj:`joblib.parallel_backend` context. ``-1`` means using all processors.

    Examples
    --------
    >>> import pandas as pd
    >>> from giottotime.feature_creation import BettiCurvesFeature
    >>> X = pd.DataFrame(range(0, 15))
    >>> betti_feature = BettiCurvesFeature()
    >>> betti_feature.transform(X)
        BettiCurvesFeature_0  BettiCurvesFeature_1  BettiCurvesFeature_2
    0                   49.0                   0.0                   0.0
    1                   49.0                   0.0                   0.0
    2                   49.0                   0.0                   0.0
    3                   49.0                   0.0                   0.0
    4                   49.0                   0.0                   0.0
    5                   49.0                   0.0                   0.0
    6                   49.0                   0.0                   0.0
    7                   49.0                   0.0                   0.0
    8                   49.0                   0.0                   0.0
    9                   49.0                   0.0                   0.0
    10                  49.0                   0.0                   0.0
    11                  49.0                   0.0                   0.0
    12                  49.0                   0.0                   0.0
    13                  49.0                   0.0                   0.0
    14                  49.0                   0.0                   0.0

    """

    def __init__(
        self,
        betti_mode: str = "mean",
        output_name: str = "BettiCurvesFeature",
        betti_homology_dimensions: Iterable = (0, 1, 2),
        betti_n_values: int = 100,
        betti_rolling: int = 1,
        betti_n_jobs: Optional[int] = None,
        takens_parameters_type: str = "search",
        takens_dimension: int = 5,
        takens_stride: int = 1,
        takens_time_delay: int = 1,
        takens_n_jobs: Optional[int] = 1,
        sliding_window_width: int = 10,
        sliding_stride: int = 1,
        diags_metric: Union[str, Callable] = "euclidean",
        diags_coeff: int = 2,
        diags_max_edge_length: float = np.inf,
        diags_homology_dimensions: Iterable = (0, 1, 2),
        diags_infinity_values: Optional[float] = None,
        diags_n_jobs: Optional[int] = 1,
    ):
        super().__init__(
            output_name=output_name,
            takens_parameters_type=takens_parameters_type,
            takens_dimension=takens_dimension,
            takens_stride=takens_stride,
            takens_time_delay=takens_time_delay,
            takens_n_jobs=takens_n_jobs,
            sliding_window_width=sliding_window_width,
            sliding_stride=sliding_stride,
            diags_metric=diags_metric,
            diags_coeff=diags_coeff,
            diags_max_edge_length=diags_max_edge_length,
            diags_homology_dimensions=diags_homology_dimensions,
            diags_infinity_values=diags_infinity_values,
            diags_n_jobs=diags_n_jobs,
        )
        self.betti_mode = betti_mode
        self.betti_homology_dimensions = betti_homology_dimensions
        self.betti_n_values = betti_n_values
        self.betti_n_jobs = betti_n_jobs
        self.betti_rolling = betti_rolling

    def transform(self, time_series: pd.DataFrame) -> pd.DataFrame:
        """From the initial DataFrame ``time_series``, compute the persistence diagrams
        and detect the average lifetime for a given homology dimension. Then, assign a
        value to each initial data points.

        Parameters
        ----------
        time_series : pd.DataFrame, shape (n_samples, 1), required
            The DataFrame on which to compute the feature_creation.

        Returns
        -------
        time_series_t : pd.DataFrame, shape (n_samples, 1)
            A DataFrame containing, for each original data-point, the average lifetime
            associated to it. If, given the initial parameters, a point was excluded
            from the computation, its value is set to ``Nan``.

        """
        persistence_diagrams = self._compute_persistence_diagrams(time_series)
        betti_curves = self._compute_betti_curves(persistence_diagrams)

        betti_features = self._compute_betti_features(betti_curves)

        output_dfs = []
        for betti_feature in betti_features:
            original_points = self._compute_n_points(len(betti_feature))

            output_dfs.append(
                _align_indices(time_series, original_points, betti_feature)
            )

        time_series_aligned = pd.concat(output_dfs, axis=1)
        time_series_t = self._rename_columns(time_series_aligned)

        return time_series_t

    def _compute_betti_curves(self, diagrams: np.ndarray) -> List:
        betti_curves = diag.BettiCurve()
        betti_curves.fit(diagrams)
        X_betti_curves = betti_curves.transform(diagrams)

        betti_curves = []
        for h_dim in self.betti_homology_dimensions:
            betti_curves.append(pd.DataFrame(X_betti_curves[:, h_dim, :]))

        return betti_curves

    def _compute_betti_features(
        self, betti_curves: List[pd.DataFrame]
    ) -> List[np.ndarray]:
        if self.betti_mode == "mean":
            betti_features = self._compute_betti_mean(betti_curves)

        elif self.betti_mode == "arg_max":
            betti_features = self._compute_arg_max_by_time(betti_curves)

        else:
            raise ValueError(
                f"The valid values for 'betti_mode' are 'mean' "
                f"or 'arg_max', instead has value "
                f"{self.betti_mode}."
            )

        return betti_features

    def _compute_betti_mean(
        self, betti_surfaces: List[pd.DataFrame]
    ) -> List[pd.DataFrame]:
        betti_means = []
        for betti_surface in betti_surfaces:
            betti_means.append(
                betti_surface.groupby(betti_surface.index)
                .apply(lambda g: _find_mean_nonzero(g))
                .rolling(self.betti_rolling)
                .mean()
                .values
            )

        return betti_means

    def _compute_arg_max_by_time(
        self, betti_surfaces: List[pd.DataFrame]
    ) -> List[np.ndarray]:
        betti_arg_maxes = []
        for betti_surface in betti_surfaces:
            arg_max = np.argmax(np.array(betti_surface), axis=1)
            betti_arg_maxes.append(arg_max)

        return betti_arg_maxes



"""
The :mod:`giottotime.causality_tests` module deals with the causality tests for time
series data.
"""

from .base import CausalityTest
from .shifted_linear_coefficient import ShiftedLinearCoefficient
from .shifted_pearson_correlation import ShiftedPearsonCorrelation


__all__ = [
    "CausalityTest",
    "ShiftedLinearCoefficient",
    "ShiftedPearsonCorrelation",
]



from typing import Iterable, Dict, Optional, Union, Callable

import giotto.diagrams as diag
import numpy as np
import pandas as pd

from .base import TDAFeatures, _align_indices


class AmplitudeFeature(TDAFeatures):
    """Compute the list of average lifetime for each time window, starting from the
    persistence diagrams.

    Parameters
    ----------
    metric : ``'bottleneck'`` | ``'wasserstein'`` | ``'landscape'`` | \
        ``'betti'`` | ``'heat'``, optional, (default=``'landscape'``)
        Distance or dissimilarity function used to define the amplitude of a subdiagram
        as its distance from the diagonal diagram:
        - ``'bottleneck'`` and ``'wasserstein'`` refer to the identically named
          perfect-matching--based notions of distance.
        - ``'landscape'`` refers to the :math:`L^p` distance between persistence
          landscapes.
        - ``'betti'`` refers to the :math:`L^p` distance between Betti curves.
        - ``'heat'`` refers to the :math:`L^p` distance between Gaussian-smoothed
          diagrams.

    output_name: str, optional, default: ``'AmplitudeFeature'``
        The name of the output column.

    amplitude_metric_params : Dict, optional, default: ``None``
        Additional keyword arguments for the metric function:
        - If ``metric == 'bottleneck'`` there are no available arguments.
        - If ``metric == 'wasserstein'`` the only argument is `p` (int,
          default: ``2``).
        - If ``metric == 'betti'`` the available arguments are `p` (float,
          default: ``2.``) and `n_values` (int, default: ``100``).
        - If ``metric == 'landscape'`` the available arguments are `p`
          (float, default: ``2.``), `n_values` (int, default: ``100``) and
          `n_layers` (int, default: ``1``).
        - If ``metric == 'heat'`` the available arguments are `p` (float,
          default: ``2.``), `sigma` (float, default: ``1.``) and `n_values`
          (int, default: ``100``).

    amplitude_order : float, optional, default: ``2.``
        If ``None``, :meth:`transform` returns for each diagram a vector of amplitudes
        corresponding to the dimensions in :attr:`homology_dimensions_`. Otherwise, the
        :math:`p`-norm of these vectors with :math:`p` equal to `order` is taken.

    amplitude_n_jobs : int, optional, default: ``None``
        The number of jobs to use for the computation. ``None`` means 1 unless in a
        :obj:`joblib.parallel_backend` context. ``-1`` means using all processors.

    takens_parameters_type: ``'search'`` | ``'fixed'``, optional, default: ``'search'``
        If set to ``'fixed'``, the values of `time_delay` and `dimension are used
        directly in :meth:`transform`. If set to ``'search'``, those values are only
        used as upper bounds in a search as follows: first, an optimal time delay is
        found by minimising the time delayed mutual information; then, a heuristic based
        on an algorithm in [2]_ is used to select an embedding dimension which, when
        increased, does not reveal a large proportion of "false nearest neighbors".

    takens_time_delay : int, optional, default: ``1``
        Time delay between two consecutive values for constructing one embedded point.
        If `parameters_type` is ``'search'``, it corresponds to the maximal embedding
        time delay that will be considered.

    takens_dimension : int, optional, default: ``5``
        Dimension of the embedding space. If `parameters_type` is ``'search'``, it
        corresponds to the maximum embedding dimension that will be considered.

    takens_stride : int, optional, default: ``1``
        Stride duration between two consecutive embedded points. It defaults to 1 as
        this is the usual value in the statement of Takens's embedding theorem.

    takens_n_jobs : int: , optional, default: ``None``
        The number of jobs to use for the computation. ``None`` means 1 unless in a
        :obj:`joblib.parallel_backend` context. ``-1`` means using all processors.

    sliding_window_width : int, optional, default: ``10``
        Width of each sliding window. Each window contains ``width + 1`` objects from
        the original time series.

    sliding_stride : int, optional, default: ``1``
        Stride between consecutive windows.

    diags_metric : Union[str, Callable], optional, default: ``'euclidean'``
        If set to `'precomputed'`, input data is to be interpreted as a collection of
        distance matrices. Otherwise, input data is to be interpreted as a collection
        of point clouds (i.e. feature arrays), and `metric` determines a rule with which
        to calculate distances between pairs of instances (i.e. rows) in these arrays.
        If `metric` is a string, it must be one of the options allowed by
        :obj:`scipy.spatial.distance.pdist` for its metric parameter, or a metric listed
        in :obj:`sklearn.pairwise.PAIRWISE_DISTANCE_FUNCTIONS`, including "euclidean",
        "manhattan", or "cosine". If `metric` is a callable function, it is called on
        each pair of instances and the resulting value recorded. The callable should
        take two arrays from the entry in `X` as input, and return a value indicating
        the distance between them.

    diags_homology_dimensions : Iterable, optional, default: ``(0, 1)``
        Dimensions (non-negative integers) of the topological feature_creation to be
        detected.

    diags_coeff : int prime, optional, default: ``2``
        Compute homology with coefficients in the prime field
        :math:`\\mathbb{F}_p = \\{ 0, \\ldots, p - 1 \\}` where
        :math:`p` equals `coeff`.

    diags_max_edge_length : float, optional, default: ``np.inf``
        Upper bound on the maximum value of the Vietoris-Rips filtration parameter.
        Points whose distance is greater than this value will never be connected by an
        edge, and topological feature_creation at scales larger than this value will not
        be detected.

    diags_infinity_values : float, optional, default: ``None``
        Which death value to assign to feature_creation which are still alive at
        filtration value `max_edge_length`. ``None`` has the same behaviour as
        `max_edge_length`.

    diags_n_jobs : int, optional, default: ``None``
        The number of jobs to use for the computation. ``None`` means 1 unless in a
        :obj:`joblib.parallel_backend` context. ``-1`` means using all processors.

    Examples
    --------
    >>> import pandas as pd
    >>> from giottotime.feature_creation import AmplitudeFeature
    >>> X = pd.DataFrame(range(0, 15))
    >>> ampl_feature = AmplitudeFeature()
    >>> ampl_feature.transform(X)
        AmplitudeFeature
    0           0.485467
    1           0.485467
    2           0.485467
    3           0.485467
    4           0.485467
    5           0.485467
    6           0.485467
    7           0.485467
    8           0.485467
    9           0.485467
    10          0.485467
    11          0.485467
    12          0.485467
    13          0.485467
    14          0.485467

    """

    def __init__(
        self,
        metric: str = "landscape",
        output_name: str = "AmplitudeFeature",
        amplitude_metric_params: Optional[Dict] = None,
        amplitude_order: Dict = 2,
        amplitude_n_jobs: Optional[float] = None,
        takens_parameters_type: str = "search",
        takens_dimension: int = 5,
        takens_stride: int = 1,
        takens_time_delay: int = 1,
        takens_n_jobs: Optional[int] = 1,
        sliding_window_width: int = 10,
        sliding_stride: int = 1,
        diags_metric: Union[str, Callable] = "euclidean",
        diags_coeff: int = 2,
        diags_max_edge_length: float = np.inf,
        diags_homology_dimensions: Iterable = (0, 1, 2),
        diags_infinity_values: Optional[float] = None,
        diags_n_jobs: Optional[int] = 1,
    ):
        super().__init__(
            output_name=output_name,
            takens_parameters_type=takens_parameters_type,
            takens_dimension=takens_dimension,
            takens_stride=takens_stride,
            takens_time_delay=takens_time_delay,
            takens_n_jobs=takens_n_jobs,
            sliding_window_width=sliding_window_width,
            sliding_stride=sliding_stride,
            diags_metric=diags_metric,
            diags_coeff=diags_coeff,
            diags_max_edge_length=diags_max_edge_length,
            diags_homology_dimensions=diags_homology_dimensions,
            diags_infinity_values=diags_infinity_values,
            diags_n_jobs=diags_n_jobs,
        )
        self.metric = metric
        self.amplitude_metric_params = amplitude_metric_params
        self.amplitude_order = amplitude_order
        self.amplitude_n_jobs = amplitude_n_jobs

    def transform(self, time_series: pd.DataFrame) -> pd.DataFrame:
        """From the initial DataFrame ``time_series``, compute the persistence diagrams
        and detect the average lifetime for a given homology dimension. Then, assign a
        value to each initial data points.

        Parameters
        ----------
        time_series : pd.DataFrame, shape (n_samples, 1), required
            The DataFrame on which to compute the feature_creation.

        Returns
        -------
        time_series_t : pd.DataFrame, shape (n_samples, 1)
            A DataFrame containing, for each original data-point, the average lifetime
            associated to it. If, given the initial parameters, a point was excluded
            from the computation, its value is set to ``Nan``.

        """
        persistence_diagrams = self._compute_persistence_diagrams(time_series)
        amplitudes = self._calculate_amplitude_feature(persistence_diagrams)

        original_points = self._compute_n_points(len(amplitudes))

        time_series_aligned = _align_indices(time_series, original_points, amplitudes)
        time_series_t = self._rename_columns(time_series_aligned)

        return time_series_t

    def _calculate_amplitude_feature(self, diagrams: np.ndarray) -> np.ndarray:
        amplitude = diag.Amplitude(
            metric=self.metric,
            order=self.amplitude_order,
            metric_params=self.amplitude_metric_params,
            n_jobs=self.amplitude_n_jobs,
        )
        return amplitude.fit_transform(diagrams)



from typing import Callable

from sklearn.metrics import mean_squared_error
from scipy.optimize import minimize

import numpy as np
import pandas as pd
from sklearn.utils.validation import check_is_fitted

from .base import TrendModel


class PolynomialTrend(TrendModel):
    """A model for fitting, predicting and removing an polynomial trend from a time
    series.

    Parameters
    ----------
    order : int, required
        The order of the polynomial.

    loss : Callable, optional, default: ``mean_squared_error``
        The loss function to use when fitting the model. The loss function must
         accept y_true, y_pred and return a single real number.

    method : str, optional, default: ``'BFGS``
        The method to use in order to minimize the loss function.

    """

    def __init__(
        self, order: int = 2, loss: Callable = mean_squared_error, method: str = "BFGS"
    ):
        self.order = order
        self.loss = loss
        self.method = method

    def fit(self, ts: pd.DataFrame,) -> "PolynomialTrend":
        """Fit the model on the ``time_series``, with respect to the provided ``loss``
        and using the provided ``method``. In order to see which methods are available,
        please check the 'scipy' `documentation
        <https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.minimize.html>`_.

        Parameters
        ----------
        ts: pd.DataFrame, shape (n_samples, n_features), required
            The time series on which to fit the model.

        Returns
        -------
        self : TrendModel
            The fitted object.

        """

        def prediction_loss(weights: np.ndarray) -> float:
            p = np.poly1d(weights)
            predictions = [p(t) for t in range(0, ts.shape[0])]
            return self.loss(ts.values, predictions)

        model_weights = np.zeros(self.order + 1)

        res = minimize(
            prediction_loss, model_weights, method=self.method, options={"disp": False}
        )

        self.model_weights_ = res["x"]
        self.t0_ = ts.index[0]
        freq = ts.index.freq

        if freq is not None:
            self.period_ = freq
        else:
            self.period_ = ts.index[1] - ts.index[0]

        return self

    def predict(self, ts: pd.DataFrame) -> pd.DataFrame:
        """Using the fitted polynomial, predict the values starting from ``X``.

        Parameters
        ----------
        ts: pd.DataFrame, shape (n_samples, 1), required
            The time series on which to predict.

        Returns
        -------
        predictions : pd.DataFrame, shape (n_samples, 1)
            The output predictions.

        Raises
        ------
        NotFittedError
            Raised if the model is not fitted yet.

        """
        check_is_fitted(self)

        p = np.poly1d(self.model_weights_)
        predictions = p(ts.values)
        return predictions

    def transform(self, ts: pd.DataFrame) -> pd.DataFrame:
        """Transform the ``time_series`` by removing the trend.

        Parameters
        ----------
        ts: pd.DataFrame, shape (n_samples, 1), required
            The time series to transform.

        Returns
        -------
        ts_t : pd.DataFrame, shape (n_samples, n_features)
            The transformed time series, without the trend.

        """
        check_is_fitted(self)

        p = np.poly1d(self.model_weights_)
        time_steps = (ts.index - self.t0_) / self.period_

        predictions = pd.Series(index=ts.index, data=[p(t) for t in time_steps])

        return ts.sub(predictions, axis=0)



from typing import List, Optional

import hypothesis.strategies as st

from giottotime.feature_creation import FeatureCreation
from giottotime.feature_creation.base import Feature
from .time_indexes import giotto_time_series


@st.composite
def X_y_matrices(
    draw,
    horizon: int,
    time_series_features: List[Feature],
    min_length: Optional[int] = None,
    allow_nan_infinity: bool = True,
):
    """ Returns a strategy that generates X and y feature matrices.

    Parameters
    ----------
    horizon : ``int``, required
        The number of steps to forecast in the future. It affects the y shape.

    time_series_features : ``List[TimeSeriesFeature]``, required
        The list of TimeSeriesFeature that is given as input to the FeaturesCreation

    min_length : ``int``, optional, (default=``None``)
        Minimum length of the matrices

    allow_nan_infinity : ``bool``, optional, (default=``True``)
        Allow nan and infinity in the starting time series

    Returns
    -------
    X : pd.DataFrame
        X feature matrix

    y : pd.DataFrame
        y feature matrix
    """
    min_length = min_length if min_length is not None else 1
    period_index_series = draw(
        giotto_time_series(
            min_length=min_length,
            allow_nan=allow_nan_infinity,
            allow_infinity=allow_nan_infinity,
        )
    )
    feature_creation = FeatureCreation(
        horizon=horizon, time_series_features=time_series_features
    )
    X, y = feature_creation.fit_transform(period_index_series)
    return X, y


@st.composite
def y_matrices(
    draw,
    horizon: int = 3,
    min_length: Optional[int] = None,
    allow_nan_infinity: bool = True,
):
    """ Returns a strategy that generates the y feature matrix.

    Parameters
    ----------
    horizon : ``int``, optional, (default=3)
        the number of steps to forecast in the future. It affects the y shape.

    min_length : ``int``, optional, (default=``None``)
        minimum length of the matrices

    allow_nan_infinity : ``bool``, optional, (default=``True``)
        allow nan and infinity in the starting time series

    Returns
    -------
    y : ``pd.DataFrame``
            y feature matrix
    """
    X, y = draw(
        X_y_matrices(
            horizon=horizon,
            time_series_features=[],
            min_length=min_length,
            allow_nan_infinity=allow_nan_infinity,
        )
    )
    return y


@st.composite
def X_matrices(
    draw,
    time_series_features: List[Feature],
    min_length: Optional[int] = None,
    allow_nan_infinity: bool = True,
):
    """ Returns a strategy that generates the X feature matrix.

    Parameters
    ----------
    time_series_features : ``List[TimeSeriesFeature]``, required
        the list of TimeSeriesFeature that is given as input to the
        FeaturesCreation

    min_length : ``int``, optional, (default=``None``)
        minimum length of the matrices

    allow_nan_infinity : ``bool``, optional, (default=``True``)
        allow nan and infinity in the starting time series

    Returns
    -------
    X : ``pd.DataFrame``
            X feature matrix
    """
    X, y = draw(
        X_y_matrices(
            horizon=1,
            time_series_features=time_series_features,
            min_length=min_length,
            allow_nan_infinity=allow_nan_infinity,
        )
    )
    return X



from typing import Tuple

import pandas as pd
from hypothesis import given

from giottotime.feature_creation import ShiftFeature, MovingAverageFeature
from giottotime.utils.hypothesis.feature_matrices import (
    X_y_matrices,
    X_matrices,
    y_matrices,
)

features = [
    ShiftFeature(0, "0"),
    ShiftFeature(1, "1"),
    MovingAverageFeature(3, "3"),
]


class TestXyMatrices:
    @given(X_y_matrices(horizon=1, time_series_features=features))
    def test_horizon_1(self, X_y: Tuple[pd.DataFrame, pd.DataFrame]):
        X, y = X_y
        assert y.shape[1] == 1

    @given(X_y_matrices(horizon=2, time_series_features=features))
    def test_horizon_2(self, X_y: Tuple[pd.DataFrame, pd.DataFrame]):
        X, y = X_y
        assert y.shape[1] == 2

    @given(X_y_matrices(horizon=3, time_series_features=features))
    def test_horizon_3(self, X_y: Tuple[pd.DataFrame, pd.DataFrame]):
        X, y = X_y
        assert y.shape[1] == 3

    @given(X_y_matrices(horizon=3, time_series_features=features))
    def test_X_column_names(self, X_y: Tuple[pd.DataFrame, pd.DataFrame]):
        X, y = X_y
        assert list(X.columns) == ["0", "1", "3"]

    @given(X_y_matrices(horizon=3, time_series_features=features))
    def test_X_shape_correct(self, X_y: Tuple[pd.DataFrame, pd.DataFrame]):
        X, y = X_y
        assert X.shape[1] == len(features)

    @given(X_y_matrices(horizon=3, time_series_features=features))
    def test_y_column_names(self, X_y: Tuple[pd.DataFrame, pd.DataFrame]):
        X, y = X_y
        assert list(y.columns) == ["y_1", "y_2", "y_3"]

    @given(X_y_matrices(horizon=3, time_series_features=features))
    def test_shape_consistent(self, X_y: Tuple[pd.DataFrame, pd.DataFrame]):
        X, y = X_y
        assert X.shape[0] == y.shape[0]

    @given(
        X_y_matrices(horizon=3, time_series_features=features, allow_nan_infinity=False)
    )
    def test_allow_nan_false(self, X_y: Tuple[pd.DataFrame, pd.DataFrame]):
        X, y = X_y
        assert X.dropna(axis=0, how="any").shape[0] == max(X.shape[0] - 2, 0)
        assert y.dropna(axis=0, how="any").shape[0] == max(y.shape[0] - 3, 0)


class TestXMatrices:
    @given(X_matrices(time_series_features=features))
    def test_X_shape_correct(self, X: pd.DataFrame):
        assert X.shape[1] == len(features)

    @given(X_matrices(time_series_features=features))
    def test_X_column_names(self, X: pd.DataFrame):
        assert list(X.columns) == ["0", "1", "3"]

    @given(X_matrices(time_series_features=features, allow_nan_infinity=False))
    def test_allow_nan_false(self, X: pd.DataFrame):
        assert X.dropna(axis=0, how="any").shape[0] == max(X.shape[0] - 2, 0)


class TestYMatrices:
    @given(y_matrices(horizon=1))
    def test_horizon_1(self, y: pd.DataFrame):
        assert y.shape[1] == 1

    @given(y_matrices(horizon=2))
    def test_horizon_2(self, y: pd.DataFrame):
        assert y.shape[1] == 2

    @given(y_matrices(horizon=3))
    def test_horizon_3(self, y: pd.DataFrame):
        assert y.shape[1] == 3

    @given(y_matrices(horizon=3))
    def test_y_column_names(self, y: pd.DataFrame):
        assert list(y.columns) == ["y_1", "y_2", "y_3"]

    @given(y_matrices(horizon=3, allow_nan_infinity=False))
    def test_allow_nan_false(self, y: pd.DataFrame):
        assert y.dropna(axis=0, how="any").shape[0] == max(y.shape[0] - 3, 0)



from random import random

import numpy as np
import pandas as pd
from hypothesis import given, settings
from hypothesis.extra.numpy import arrays
from hypothesis.strategies import floats

from giottotime.models.regressors.linear_regressor import LinearRegressor


class TestLinearRegressor:
    def test_linear_regressor(self):
        train, test = train_test_dataframe()

        predictions = compute_predictions_for_train_test(train, test)
        expected = compute_expectation_from_test(test)

        np.testing.assert_array_almost_equal(predictions, expected, decimal=2)

    @settings(deadline=None)
    @given(arrays(dtype=float, shape=(100, 1), elements=floats(-10000000, 10000000)))
    def test_linear_regressor_random_array(self, random_array):
        train, test = train_test_dataframe(random_array)

        predictions = compute_predictions_for_train_test(train, test)
        expected = compute_expectation_from_test(test)

        np.testing.assert_array_almost_equal(predictions, expected, decimal=0)


def train_test_dataframe(
    random_array: np.ndarray = None,
) -> (pd.DataFrame, pd.DataFrame):
    random_array = (
        random_array if random_array is not None else [random() for _ in range(100)]
    )

    a1, a2, b = random() * 10, random() * 100, 2 * (1 - random())

    df = pd.DataFrame()
    df["x1"] = list(range(100))
    df["x2"] = random_array
    df["y"] = [b + a1 * t for t in range(100)]
    df["y"] = df["y"] + a2 * df["x2"]

    train = df[:90]
    test = df[90:]

    return train, test


def compute_predictions_for_train_test(
    train: pd.DataFrame, test: pd.DataFrame
) -> np.ndarray:
    lr = LinearRegressor()

    lr.fit(train[["x1", "x2"]], train["y"], x0=[0, 0, 0])

    preds_y = lr.predict(test[["x1", "x2"]])
    preds_y = preds_y / np.sum(preds_y)

    return preds_y


def compute_expectation_from_test(test: pd.DataFrame) -> np.ndarray:
    test_y = test["y"].values
    test_y = test_y / np.sum(test_y)
    return test_y



from abc import ABC
from typing import Callable

from sklearn.metrics import mean_squared_error
import pandas as pd

from .base import IndexDependentFeature
from giottotime.models.trend_models import TrendModel
from giottotime.models.trend_models import PolynomialTrend
from giottotime.models.trend_models import ExponentialTrend

__all__ = [
    "DetrendedFeature",
    "RemovePolynomialTrend",
    "RemoveExponentialTrend",
]


class DetrendedFeature(IndexDependentFeature):
    """Apply a de-trend transformation to a time series.

    Parameters
    ----------
    trend_model : TrendModel, optional, default: ``PolynomialTrend()``
        The kind of trend removal to apply.

    output_name : str, optional, default: ``'DetrendedFeature'``
        The name of the output column.

    Examples
    --------
    >>> import pandas as pd
    >>> from giottotime.feature_creation import DetrendedFeature
    >>> from giottotime.models import PolynomialTrend
    >>> model = PolynomialTrend(order=2)
    >>> detrend_feature = DetrendedFeature(trend_model=model)
    >>> time_index = pd.date_range("2020-01-01", "2020-01-10")
    >>> X = pd.DataFrame(range(0, 10), index=time_index)
    >>> detrend_feature.transform(X)
                DetrendedFeature
    2020-01-01      2.092234e-06
    2020-01-02      6.590209e-07
    2020-01-03     -4.104701e-07
    2020-01-04     -1.116238e-06
    2020-01-05     -1.458284e-06
    2020-01-06     -1.436607e-06
    2020-01-07     -1.051207e-06
    2020-01-08     -3.020852e-07
    2020-01-09      8.107597e-07
    2020-01-10      2.287327e-06

    """

    def __init__(
        self, trend_model: TrendModel, output_name: str = "DetrendedFeature",
    ):
        super().__init__(output_name)
        self.trend_model = trend_model

    def transform(self, time_series: pd.DataFrame) -> pd.DataFrame:
        """Apply a de-trend transformation to the input ``time_series``.

        Parameters
        ----------
        time_series : pd.DataFrame, shape (n_samples, 1), required
            The time series on which to apply the de-trend transformation.

        Returns
        -------
        time_series_t : pd.DataFrame, shape (n_samples, 1)
            The de-trended time series.

        """
        self.trend_model.fit(time_series)
        time_series_t = self.trend_model.transform(time_series)
        time_series_t = self._rename_columns(time_series_t)
        return time_series_t


class RemovePolynomialTrend(DetrendedFeature):
    """Apply a de-trend transformation to a time series using a polynomial with a given
    degree.

    Parameters
    ----------
    polynomial_order : int, optional, default: ``1``
        The order of the polynomial to use when de-trending the time series.

    loss : Callable, optional, default: ``mean_squared_error``
        The function to use in order to minimize the loss.

    output_name : str, optional, default: ``'RemovePolynomialTrend'``
        The name of the output column.

    Examples
    --------
    >>> import pandas as pd
    >>> from giottotime.feature_creation import RemovePolynomialTrend
    >>> detrend_feature = RemovePolynomialTrend(polynomial_order=4)
    >>> time_index = pd.date_range("2020-01-01", "2020-01-10")
    >>> X = pd.DataFrame(range(0, 10), index=time_index)
    >>> detrend_feature.transform(X)
                RemovePolynomialTrend
    2020-01-01               0.000969
    2020-01-02              -0.001216
    2020-01-03              -0.000915
    2020-01-04               0.000201
    2020-01-05               0.001021
    2020-01-06               0.000995
    2020-01-07               0.000133
    2020-01-08              -0.000993
    2020-01-09              -0.001249
    2020-01-10               0.001059

    """

    def __init__(
        self,
        polynomial_order: int = 2,
        loss: Callable = mean_squared_error,
        output_name: str = "RemovePolynomialTrend",
    ):
        trend_model = PolynomialTrend(order=polynomial_order, loss=loss)
        super().__init__(trend_model=trend_model, output_name=output_name)


class RemoveExponentialTrend(DetrendedFeature):
    """Apply a de-trend transformation to a time series using an exponential function.

    Parameters
    ----------
    loss : Callable, optional, default: ``mean_squared_error``
        The function to use in order to minimize the loss.

    output_name : str, optional, default: ``'RemoveExponentialTrend'``
        The name of the output column.

    Examples
    --------
    >>> import pandas as pd
    >>> from giottotime.feature_creation import RemoveExponentialTrend
    >>> detrend_feature = RemoveExponentialTrend()
    >>> time_index = pd.date_range("2020-01-01", "2020-01-10")
    >>> X = pd.DataFrame(range(0, 10), index=time_index)
    >>> detrend_feature.transform(X)
                RemoveExponentialTrend
    2020-01-01               -1.000000
    2020-01-02               -0.295788
    2020-01-03                0.320933
    2020-01-04                0.824285
    2020-01-05                1.180734
    2020-01-06                1.346829
    2020-01-07                1.266264
    2020-01-08                0.866080
    2020-01-09                0.051740
    2020-01-10               -1.299262

    """

    def __init__(
        self,
        loss: Callable = mean_squared_error,
        output_name: str = "RemoveExponentialTrend",
    ):
        trend_model = ExponentialTrend(loss=loss)
        super().__init__(trend_model=trend_model, output_name=output_name)



from typing import List

import pandas as pd

from .base import Feature
from .index_dependent_features import ShiftFeature


def _check_feature_names(time_series_features: List[Feature]) -> None:
    feature_output_names = [feature.output_name for feature in time_series_features]
    if len(set(feature_output_names)) != len(feature_output_names):
        raise ValueError(
            "The input features should all have different names, instead "
            f"they are {feature_output_names}."
        )


class FeatureCreation:
    """Class responsible for the generation of the feature_creation, starting from a
    list of TimeSeriesFeature.

    Parameters
    ----------
    time_series_features : List[TimeSeriesFeature], required
        The list of ``TimeSeriesFeature`` from which to compute the feature_creation.

    horizon : int, optional, default: ``5``
        It represents how much into the future is necessary to predict. This corresponds
        to the number of shifts that are going to be performed on y.

    Examples
    --------
    >>> import pandas as pd
    >>> from giottotime.feature_creation import FeatureCreation
    >>> from giottotime.feature_creation import ShiftFeature, MovingAverageFeature
    >>> ts = pd.DataFrame([0, 1, 2, 3, 4, 5, 6, 7, 8, 9])
    >>> shift_feature = ShiftFeature(shift=1)
    >>> mv_avg_feature = MovingAverageFeature(window_size=2)
    >>> feature_creation = FeatureCreation(horizon=3,
    ...                                    time_series_features=[shift_feature,
    ...                                                          mv_avg_feature])
    >>> X, y = feature_creation.fit_transform(ts)
    >>> X
       ShiftFeature  MovingAverageFeature
    0           NaN                   NaN
    1           0.0                   0.5
    2           1.0                   1.5
    3           2.0                   2.5
    4           3.0                   3.5
    5           4.0                   4.5
    6           5.0                   5.5
    7           6.0                   6.5
    8           7.0                   7.5
    9           8.0                   8.5
    >>> y
       y_1  y_2  y_3
    0  1.0  2.0  3.0
    1  2.0  3.0  4.0
    2  3.0  4.0  5.0
    3  4.0  5.0  6.0
    4  5.0  6.0  7.0
    5  6.0  7.0  8.0
    6  7.0  8.0  9.0
    7  8.0  9.0  NaN
    8  9.0  NaN  NaN
    9  NaN  NaN  NaN
    """

    def __init__(self, time_series_features: List[Feature], horizon: int = 5):
        _check_feature_names(time_series_features)

        self.time_series_features = time_series_features
        self.horizon = horizon

    def fit(self, time_series: pd.DataFrame) -> "FeatureCreation":
        """Fit the all the features inside ``time_series_features``.

        Parameters
        ----------
        time_series : pd.DataFrame, required
            The time series on which to fit the data.

        Returns
        -------
        self: FeatureCreation
            The FeatureCreation object itself, containing the fitted features.

        """
        for time_series_feature in self.time_series_features:
            time_series_feature.fit(time_series)
        return self

    def transform(self, time_series: pd.DataFrame) -> (pd.DataFrame, pd.DataFrame):
        """Create the X matrix by generating the feature_creation, starting from the
        original ``time_series`` and using the list of ``time_series_features``. Also
        create the y matrix, by generating ``horizon`` number of shifts of the
         ``time_series``.

        Parameters
        ----------
        time_series : pd.DataFrame, shape (n_samples, 1), required
            The time series on which to compute the ``X`` and ``y`` matrices.

        Returns
        -------
        X, y: (pd.DataFrame, pd.DataFrame), shape ((n_samples, n_features), \
            n_samples, horizon))
            A tuple containing the ``X`` and ``y`` matrices.

        """
        X = self._create_x_features(time_series)
        y = self._create_y_shifts(time_series)
        return X, y

    def _create_y_shifts(self, time_series: pd.DataFrame) -> pd.DataFrame:
        y = pd.DataFrame(index=time_series.index)
        for k in range(1, self.horizon + 1):
            shift_feature = ShiftFeature(-k, f"shift_{k}")
            y[f"y_{k}"] = shift_feature.transform(time_series)

        return y

    def _create_x_features(self, time_series: pd.DataFrame) -> pd.DataFrame:
        features = pd.DataFrame(index=time_series.index)
        for time_series_feature in self.time_series_features:
            x_transformed = time_series_feature.transform(time_series)
            features = pd.concat([features, x_transformed], axis=1)

        return features

    def fit_transform(self, time_series: pd.DataFrame) -> (pd.DataFrame, pd.DataFrame):
        """Fit and then transform on the same time series.

        Parameters
        ----------
        time_series : pd.DataFrame, required
            The time series on which to fit and to transform.

        Returns
        -------
        X, y: (pd.DataFrame, pd.DataFrame), shape ((n_samples, n_features), \
            n_samples, horizon))
            A tuple containing the ``X`` and ``y`` matrices.

        """
        self.fit(time_series)
        X, y = self.transform(time_series)
        return X, y



import importlib
from typing import Optional, Union, List

import numpy as np
import pandas as pd
import workalendar

from .base import IndexDependentFeature


def check_index(time_series: pd.DataFrame) -> None:
    index_series = pd.Series(time_series.index)
    unique_differences = pd.unique(index_series.diff().dropna().values)

    if len(unique_differences) != 1:
        raise ValueError(
            "The time series should be evenly spaced in order to obtain "
            "meaningful results."
        )

    if not isinstance(time_series.index, pd.PeriodIndex):
        raise ValueError(
            "The input time series should have a index of type "
            f"PeriodIndex, got {type(time_series.index)} instead."
        )


class CalendarFeature(IndexDependentFeature):
    """Create a feature based on the national holidays of a specific country, based on
    a given kernel (if provided). The interface for this is based on the one of
    'workalendar'. To see which regions and countries are available, check the
    'workalendar' `documentation <https://peopledoc.github.io/workalendar/>`_.

    Parameters
    ----------
    region : str, optional, default: ``'america'``
        The region in which the ``country`` is located.

    country : str, optional, default: ``'Brazil'``
        The name of the country from which to retrieve the holidays. The country must be
         located in the given ``region``.

    start_date : str, optional, default: ``'01/01/2019'``
        The date starting from which to retrieve the holidays.

    end_date : str, optional, default: ``'01/01/2020'``
        The date until which to retrieve the holidays.

    kernel : list or np.ndarray, optional, default: ``None``
        The kernel to use when creating the feature.

    reindex_method : str, optional, default: ``pad``
        Used only if X is passed in the ``transform`` method. It is used as the method
        with which to reindex the holiday events with the index of X. This method
        should be compatible with the reindex methods provided by pandas. Please refer
        to the pandas `documentation
        https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.reindex.html`_
        for further details.

    output_name : str, optional, default: ``'CalendarFeature'``
        The name of the output column.

    Examples
    --------
    >>> from giottotime.feature_creation import CalendarFeature
    >>> cal_feature = CalendarFeature(region="europe", country="Italy", kernel=[3, 2])
    >>> cal_feature.transform()
                CalendarFeature
    2018-01-01              2.0
    2018-01-02              3.0
    2018-01-03              0.0
    2018-01-04              0.0
    2018-01-05              0.0
    ...                     ...
    2019-12-28              0.0
    2019-12-29              0.0
    2019-12-30              0.0
    2019-12-31              0.0
    2020-01-01              2.0
    """

    def __init__(
        self,
        region: str = "america",
        country: str = "Brazil",
        start_date: str = "01/01/2018",
        end_date: str = "01/01/2020",
        kernel: Union[List, np.ndarray] = None,
        reindex_method: str = "pad",
        output_name: str = "CalendarFeature",
    ):
        super().__init__(output_name)
        self.region = region
        self.country = country
        self.start_date = start_date
        self.end_date = end_date
        self.reindex_method = reindex_method

        if kernel is None or len(kernel) == 0 or not np.isfinite(kernel).all():
            raise ValueError(
                "The kernel should be an array-like object, with at least 1 element "
                f"and should only contains finite values, got {kernel} instead."
            )
        self.kernel = kernel

    def transform(self, time_series: Optional[pd.DataFrame] = None) -> pd.DataFrame:
        """Generate a DataFrame containing the events associated to the holidays of the
        selected ``country``.

        Parameters
        ----------
        time_series : pd.DataFrame, shape (n_samples, 1), optional, default: ``None``
            If provided, both ``start_date`` and ``end_date`` are going to be
            overwritten with the start and end date of the index of ``time_series``.
            Also, if provided the output DataFrame is going to be re-indexed with the
            index of ``time_series``, using the chosen ``reindex_method``.

        Returns
        -------
        events : pd.DataFrame, shape (length, 1)
            A DataFrame containing the events.

        """
        if time_series is not None:
            check_index(time_series)

        self._initialize_start_end_date(time_series)

        workalendar_region = importlib.import_module(f".{self.region}", "workalendar")
        workalendar_country = getattr(workalendar_region, self.country)()

        events = self._get_holiday_events(workalendar_country)
        if self.kernel is not None:
            events = self._apply_kernel(events)

        events_renamed = self._rename_columns(events)
        aligned_events = self._align_event_indices(time_series, events_renamed)

        return aligned_events

    def _initialize_start_end_date(self, X: pd.DataFrame):
        if X is not None:
            self.start_ = X.index.values[0]
            self.end_ = X.index.values[-1]
        else:
            self.start_ = pd.Timestamp(self.start_date)
            self.end_ = pd.Timestamp(self.end_date)

        slack_period = len(self.kernel)
        slack_days = np.timedelta64(slack_period, "D")

        slacked_start_date_ = self.start_ - slack_days
        slacked_end_date_ = self.end_ + slack_days

        self.slacked_start_date_ = pd.datetime(
            year=slacked_start_date_.year,
            month=slacked_start_date_.month,
            day=slacked_start_date_.day,
        )
        self.slacked_end_date_ = pd.datetime(
            year=slacked_end_date_.year,
            month=slacked_end_date_.month,
            day=slacked_end_date_.day,
        )

    def _get_holiday_events(self, country_mod):
        index = pd.date_range(
            start=self.slacked_start_date_, end=self.slacked_end_date_, freq="D"
        )
        years = index.year.unique()
        events = pd.DataFrame()
        for year in years:
            events = events.append(country_mod.holidays(year))

        events = events.rename(columns={0: "date", 1: "events"})
        events["date"] = pd.to_datetime(events["date"])
        events["status"] = 1
        events = events.set_index("date")
        events = self._group_by_event_name(events)

        events = events.reindex(index)
        events = events.drop(columns=["events"])

        events["status"] = events["status"].fillna(0).astype(int)
        events = events.sort_index()

        return events

    def _group_by_event_name(self, events_df):
        def aggregate_events_same_day(event):
            return pd.Series(
                {"status": 1, "events": " & ".join(event["events"].tolist())}
            )

        grouped_events = events_df.groupby(events_df.index).apply(
            aggregate_events_same_day
        )
        return grouped_events

    def _apply_kernel(self, events):
        kernel_events = events.copy()
        klen = len(self.kernel)

        def apply_kernel(w):
            if sum(w) != 0:
                return np.dot(w, self.kernel) / sum(w)
            else:
                return 0

        kernel_events["status"] = (
            kernel_events["status"]
            .rolling(klen, center=True)
            .apply(apply_kernel, raw=False)
        )
        return kernel_events

    def _align_event_indices(self, ts, events):
        if ts is not None:
            X = ts.copy()
            new_x_line = pd.DataFrame(
                columns=X.columns, index=pd.PeriodIndex([X.index[-1] + 1])
            )
            X_to_cut = pd.concat([X, new_x_line], axis=0)
            bins = pd.cut(events.index, X_to_cut.index.to_timestamp())

            grouped_events = events.groupby(bins).mean().ffill()

            grouped_events.index = pd.to_datetime(
                grouped_events.index.map(lambda row_index: row_index.left)
            )
            X.index = X.index.to_timestamp()

            events_renamed = pd.merge(
                X, grouped_events, left_index=True, right_index=True, how="inner"
            )
            events_renamed.index = ts.index
        else:
            events_renamed = events[self.start_ : self.end_]

        return events_renamed



from itertools import product

import numpy as np
import pandas as pd
from sklearn.linear_model import LinearRegression


class GrangerTest:
    def __init__(self, data):
        self.data = data

    def shifted_partial_corr_matrix(self, max_shift, x="x", y="y"):
        shifts = pd.DataFrame()

        for shift in range(max_shift):
            shifts[shift] = self.data[x].shift(shift)

        shifts = shifts.dropna()
        self.shifted_corrs = shifts.corrwith(self.data[y])

        return self.shifted_corrs

    def all_best_corr_shifts(self, max_shift):
        self.best_shifts = pd.DataFrame()

        for x, y in product(self.data.columns, repeat=2):
            res = self.get_max_corr_shift(max_shift, x=x, y=y)

            best_shift = res[1]
            max_corr = res[0]
            N = self.data.shape[0] - max_shift

            self.best_shifts = self.best_shifts.append(
                {"x": x, "y": y, "shift": best_shift, "max_corr": max_corr},
                ignore_index=True,
            )

        best_shifts = pd.pivot_table(
            self.best_shifts, index=["x"], columns=["y"], values="shift"
        )
        max_corrs = pd.pivot_table(
            self.best_shifts, index=["x"], columns=["y"], values="max_corr"
        )

        return best_shifts, max_corrs

    def get_max_corr_shift(self, max_shift, x="x", y="y"):
        shifts = pd.DataFrame()

        for shift in range(max_shift):
            shifts[shift] = self.data[x].shift(shift)

        shifts = shifts.dropna()
        self.shifted_corrs = shifts.corrwith(self.data[y])

        q = self.shifted_corrs.max(), self.shifted_corrs.idxmax()
        return q

    def all_best_coeff_shifts(self, max_shift):
        self.best_shifts = pd.DataFrame()

        for x, y in product(self.data.columns, repeat=2):
            res = self.get_max_coeff_shift(max_shift, x=x, y=y)

            best_shift = res[1]
            max_corr = res[0]
            N = self.data.shape[0] - max_shift

            self.best_shifts = self.best_shifts.append(
                {"x": x, "y": y, "shift": best_shift, "max_corr": max_corr},
                ignore_index=True,
            )

        best_shifts = pd.pivot_table(
            self.best_shifts, index=["x"], columns=["y"], values="shift"
        )
        max_corrs = pd.pivot_table(
            self.best_shifts, index=["x"], columns=["y"], values="max_corr"
        )

        return best_shifts, max_corrs

    def get_max_coeff_shift(self, max_shift, x="x", y="y"):
        shifts = pd.DataFrame()

        shifts[x] = self.data[x]
        shifts[y] = self.data[y]

        for shift in range(max_shift):
            shifts[shift] = self.data[x].shift(shift)

        shifts = shifts.dropna()
        lf = LinearRegression().fit(shifts[range(max_shift)].values, shifts[y].values)

        q = lf.coef_.max(), np.argmax(lf.coef_)
        return q


if __name__ == "__main__":
    import pandas.util.testing as testing
    from random import randint

    testing.N, testing.K = 2000, 1

    df = testing.makeTimeDataFrame(freq="MS")
    df["B"] = df["A"].shift(8) + pd.Series(
        data=np.random.normal(0, 0.3, df.shape[0]), index=df.index
    )

    for k in range(10):
        df[f"C_{k}"] = df["A"].shift(8 + randint(10, 20)) + pd.Series(
            data=np.random.normal(0, 0.3, df.shape[0]), index=df.index
        )

    df = df.dropna()

    gt = GrangerTest(df)
    # q = gt.shifted_partial_corr_matrix(20, 'A', 'B')

    w = gt.all_best_corr_shifts(max_shift=20)
    print(w[0])
    print("\n\n±±±±±±±±±±±±±±±±±±±±±±±±±±±±±±±±±±±±±±±±±±±±±±\n\n")
    print(w[1])

    print(
        "\n\n±±±±±±±±±±±±±±±±±±±±±±±±±±±±±±±±±±±±±±±±±±±±±±±±±±±±±±±±±±±±±±±±±±±±±±±±±±±±±±±±±±±±±±±±±±±±\n\n"
    )

    q = gt.all_best_coeff_shifts(max_shift=20)

    print(q[0])
    print("\n\n±±±±±±±±±±±±±±±±±±±±±±±±±±±±±±±±±±±±±±±±±±±±±±\n\n")
    print(q[1])

    print(q[1] - w[1])
    print(q[0] == w[0])

#


#



"""
The :mod:`giottotime.feature_creation` module deals with the preparation of time series
data, such as conversion to `pandas.DataFrame` with a `PeriodIndex`.
"""

from .time_series_conversion import (
    SequenceToTimeIndexSeries,
    PandasSeriesToTimeIndexSeries,
    TimeIndexSeriesToPeriodIndexSeries,
)

from .time_series_preparation import TimeSeriesPreparation

__all__ = [
    "SequenceToTimeIndexSeries",
    "PandasSeriesToTimeIndexSeries",
    "TimeIndexSeriesToPeriodIndexSeries",
    "TimeSeriesPreparation",
]



from typing import Optional

import pandas as pd
from sklearn.preprocessing import PolynomialFeatures

from .base import IndexDependentFeature

__all__ = [
    "ShiftFeature",
    "MovingAverageFeature",
    "PolynomialFeature",
    "ExogenousFeature",
]


class ShiftFeature(IndexDependentFeature):
    """Perform a shift of a DataFrame of size equal to ``shift``.

    Parameters
    ----------
    shift : int, optional, default: ``1``
        How much to shift.

    output_name : str, optional, default: ``'ShiftFeature'``
        The name of the output column.

    Notes
    -----
    The ``shift`` parameter can also accept negative values. However, this should be
    used carefully, since if the resulting feature is used for training or testing it
    might generate a leak from the feature.

    Examples
    --------
    >>> import pandas as pd
    >>> from giottotime.feature_creation import ShiftFeature
    >>> ts = pd.DataFrame([0, 1, 2, 3, 4, 5])
    >>> shift_feature = ShiftFeature(shift=3)
    >>> shift_feature.transform(ts)
       ShiftFeature
    0           NaN
    1           NaN
    2           NaN
    3           0.0
    4           1.0
    5           2.0

    """

    def __init__(self, shift: int = 1, output_name: str = "ShiftFeature"):
        super().__init__(output_name)
        self.shift = shift

    def transform(self, time_series: pd.DataFrame) -> pd.DataFrame:
        """Create a shifted version of ``time_series``.

        Parameters
        ----------
        time_series : pd.DataFrame, shape (n_samples, 1), required
            The DataFrame to shift.

        Returns
        -------
        time_series_t : pd.DataFrame, shape (n_samples, 1)
            The shifted version of the original ``time_series``.

        """
        time_series_shifted = time_series.shift(self.shift)
        time_series_t = self._rename_columns(time_series_shifted)
        return time_series_t


class MovingAverageFeature(IndexDependentFeature):
    """For each row in ``time_series``, compute the moving average of the previous
     ``window_size`` rows. If there are not enough rows, the value is Nan.

    Parameters
    ----------
    window_size : int, optional, default: ``1``
        The number of previous points on which to compute the moving average

    output_name : str, optional, default: ``'MovingAverageFeature'``
        The name of the output column.

    Examples
    --------
    >>> import pandas as pd
    >>> from giottotime.feature_creation import MovingAverageFeature
    >>> ts = pd.DataFrame([0, 1, 2, 3, 4, 5])
    >>> mv_avg_feature = MovingAverageFeature(window_size=2)
    >>> mv_avg_feature.transform(ts)
       MovingAverageFeature
    0                   NaN
    1                   0.5
    2                   1.5
    3                   2.5
    4                   3.5
    5                   4.5

    """

    def __init__(self, window_size: int = 1, output_name: str = "MovingAverageFeature"):
        super().__init__(output_name)
        self.window_size = window_size

    def transform(self, time_series: pd.DataFrame) -> pd.DataFrame:
        """Compute the moving average, for every row of ``time_series``, of the previous
        ``window_size`` elements.

        Parameters
        ----------
        time_series : pd.DataFrame, shape (n_samples, 1), required
            The DataFrame on which to compute the rolling moving average

        Returns
        -------
        time_series_t : pd.DataFrame, shape (n_samples, 1)
            A DataFrame, with the same length as ``time_series``, containing the rolling
            moving average for each element.

        """
        time_series_mvg_avg = time_series.rolling(self.window_size).mean()
        time_series_t = self._rename_columns(time_series_mvg_avg)
        return time_series_t


class PolynomialFeature(IndexDependentFeature):
    """Compute the polynomial feature_creation, of a degree equal to the input
    ``degree``.

    Parameters
    ----------
    degree: int, optional, default: ``2``
        The degree of the polynomial feature_creation.

    output_name : str, optional, default: ``'PolynomialFeature'``
        The name of the output column.

    Examples
    --------
    >>> import pandas as pd
    >>> from giottotime.feature_creation import PolynomialFeature
    >>> ts = pd.DataFrame([0, 1, 2, 3, 4, 5])
    >>> pol_feature = PolynomialFeature(degree=3, output_name="pol")
    >>> pol_feature.transform(ts)
       pol_0  pol_1  pol_2  pol_3
    0    1.0    0.0    0.0    0.0
    1    1.0    1.0    1.0    1.0
    2    1.0    2.0    4.0    8.0
    3    1.0    3.0    9.0   27.0
    4    1.0    4.0   16.0   64.0
    5    1.0    5.0   25.0  125.0

    """

    def __init__(self, degree: int = 2, output_name: str = "PolynomialFeature"):
        super().__init__(output_name)
        self.degree = degree

    def transform(self, time_series: pd.DataFrame) -> pd.DataFrame:
        """Compute the polynomial feature_creation of ``time_series``, up to a degree
        equal to ``degree``.

        Parameters
        ----------
        time_series : pd.DataFrame, shape (n_samples, 1), required
            The input DataFrame. Used only for its index.

        Returns
        -------
        time_series_t : pd.DataFrame, shape (n_samples, 1)
            The computed polynomial feature_creation.

        """
        pol_feature = PolynomialFeatures(self.degree)
        pol_of_X_array = pol_feature.fit_transform(time_series)
        pol_of_X = pd.DataFrame(pol_of_X_array, index=time_series.index)

        time_series_t = self._rename_columns(pol_of_X)

        return time_series_t


class ExogenousFeature(IndexDependentFeature):
    """Reindex ``exogenous_time_series`` with the index of ``time_series``. To check the
    documentation of ``pandas.DataFrame.reindex`` and to see which type of
    ``method`` are available, please refer to the pandas `documentation
    <https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.reindex.html>`_.

    Parameters
    ----------
    exogenous_time_series : pd.DataFrame, shape (n_samples, 1), required
        The time series to reindex

    method : str, optional, default: ``None``
        The method used to re-index. This must be a method used by the
        ``pandas.DataFrame.reindex`` method.

    output_name : str, optional, default: ``'ExogenousFeature'``
        The name of the output column.

    Examples
    --------
    >>> import pandas as pd
    >>> from giottotime.feature_creation import ExogenousFeature
    >>> ts = pd.DataFrame([0, 1, 2, 3, 4, 5], index=[3, 4, 5, 6, 7, 8])
    >>> exog_ts = pd.DataFrame([10, 8, 1, 3, 2, 7])
    >>> exog_feature = ExogenousFeature(exog_ts)
    >>> exog_feature.transform(ts)
       ExogenousFeature
    3               3.0
    4               2.0
    5               7.0
    6               NaN
    7               NaN
    8               NaN

    >>> exog_feature = ExogenousFeature(exog_ts, method="nearest")
    >>> exog_feature.transform(ts)
       ExogenousFeature
    3                 3
    4                 2
    5                 7
    6                 7
    7                 7
    8                 7
    """

    def __init__(
        self,
        exogenous_time_series: pd.DataFrame,
        method: Optional[str] = None,
        output_name: str = "ExogenousFeature",
    ):
        super().__init__(output_name)
        self.method = method
        self.exogenous_time_series = exogenous_time_series

    def transform(self, time_series: pd.DataFrame) -> pd.DataFrame:
        """Reindex the ``exogenous_time_series`` with the index of ``time_series``.

        Parameters
        ----------
        time_series : pd.DataFrame, shape (n_samples, 1), required
            The input DataFrame. Used only for its index.

        Returns
        -------
        time_series_t :  pd.DataFrame, shape (n_samples, 1)
            The original ``exogenous_time_series``, re-indexed with the index
            of ``time_series``.

        """
        exog_feature = self.exogenous_time_series.reindex(
            index=time_series.index, method=self.method
        )
        time_series_t = self._rename_columns(exog_feature)
        return time_series_t



from abc import ABCMeta, abstractmethod

import pandas as pd

from giottotime.feature_creation.base import Feature


class IndexDependentFeature(Feature, metaclass=ABCMeta):
    """Base class for all the feature classes in this package.

    """

    def fit(self, X, y=None):
        return self

    @abstractmethod
    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        raise NotImplementedError  # to exclude from pytest coverage

    def fit_transform(self, X, y=None):
        self.fit(X, y)
        return self.transform(X)



import numpy as np
import pandas as pd
import pytest
from hypothesis import given, strategies as st

from giottotime.feature_creation.index_dependent_features.tda_features import (
    NumberOfRelevantHolesFeature,
)


@given(st.integers(min_value=3))
def test_wrong_h_dim(h_dim):
    with pytest.raises(ValueError):
        NumberOfRelevantHolesFeature(output_name="ignored", h_dim=h_dim)


@given(st.floats(max_value=0))
def test_wrong_theta(theta):
    with pytest.raises(ValueError):
        NumberOfRelevantHolesFeature(output_name="ignored", theta=theta)


def test_correct_n_holes_feature():
    np.random.seed(0)

    h_dim = 0
    theta = 0.4
    output_name = "n_holes"
    n_holes_feature = NumberOfRelevantHolesFeature(
        output_name=output_name, h_dim=h_dim, theta=theta
    )
    df = pd.DataFrame(np.random.randint(0, 100, size=(30, 1)), columns=["old_name"])

    n_holes = n_holes_feature.transform(df)

    expected_n_holes = pd.DataFrame.from_dict(
        {
            output_name: [
                9.0,
                9.0,
                10.0,
                10.0,
                9.0,
                9.0,
                10.0,
                10.0,
                10.0,
                10.0,
                10.0,
                10.0,
                10.0,
                10.0,
                10.0,
                10.0,
                10.0,
                10.0,
                10.0,
                10.0,
                10.0,
                10.0,
                10.0,
                10.0,
                10.0,
                10.0,
                10.0,
                10.0,
                10.0,
                10.0,
            ]
        }
    )

    assert np.allclose(expected_n_holes, n_holes)


def test_correct_n_relevant_holes():
    np.random.seed(0)

    h_dim = 1
    theta = 0.8
    output_name = "n_holes"
    n_holes_feature = NumberOfRelevantHolesFeature(
        output_name=output_name, h_dim=h_dim, theta=theta
    )
    df = pd.DataFrame(np.random.randint(0, 100, size=(30, 1)), columns=["old_name"])

    persistence_diagrams = n_holes_feature._compute_persistence_diagrams(df)
    n_holes = n_holes_feature._compute_num_relevant_holes(persistence_diagrams)

    expected_n_holes = [1, 1, 1, 1, 2, 2, 2, 1, 1, 1, 1, 2, 2, 2, 1, 2, 1]

    assert np.allclose(expected_n_holes, n_holes)



from random import randint
from typing import List

import numpy as np
import pandas as pd
from pandas.util import testing as testing

from giottotime.causality_tests import (
    ShiftedLinearCoefficient,
    ShiftedPearsonCorrelation,
)


class TestCausalityTest:
    def test_causality_tests_shifts_shifted_pearson_correlation(self):
        expected_shifts = [randint(2, 9) * 2 for _ in range(3)]
        df = make_df_from_expected_shifts(expected_shifts)

        shifted_test = ShiftedPearsonCorrelation(target_col="A", max_shift=20)
        shifted_test.fit(df)

        shifts = shifted_test.best_shifts_.loc["A"][1:].values
        np.testing.assert_array_equal(shifts, expected_shifts)

    def test_causality_tests_shifts_shifted_linear_coefficient(self):
        expected_shifts = [randint(2, 9) * 2 for _ in range(3)]
        df = make_df_from_expected_shifts(expected_shifts)

        shifted_test = ShiftedLinearCoefficient(target_col="A", max_shift=20)
        shifted_test.fit(df)

        shifts = shifted_test.best_shifts_.loc["A"][1:].values
        np.testing.assert_array_equal(shifts, expected_shifts)

    def test_causality_tests_shifts_shifted_pearson_correlation_transform(self):
        expected_shifts = [randint(2, 9) * 2 for _ in range(3)]
        df = make_df_from_expected_shifts(expected_shifts)

        shifted_test = ShiftedPearsonCorrelation(target_col="A", max_shift=20)
        shifted_test.fit(df)

        transformation = shifted_test.transform(df).dropna()
        expected_transformation = shift_df_from_expected_shifts(df, expected_shifts)

        np.testing.assert_array_equal(transformation, expected_transformation)

    def test_causality_tests_shifts_shifted_linear_coefficient_transform(self):
        expected_shifts = [randint(2, 9) * 2 for _ in range(3)]
        df = make_df_from_expected_shifts(expected_shifts)

        shifted_test = ShiftedLinearCoefficient(target_col="A", max_shift=20)
        shifted_test.fit(df)

        transformation = shifted_test.transform(df).dropna()
        expected_transformation = shift_df_from_expected_shifts(df, expected_shifts)

        np.testing.assert_array_equal(transformation, expected_transformation)

    def test_linear_p_values(self):
        # This test and the next one just test if the p_values on the diagonal are equal
        # to 0. Is hard to implement other unittets, since the bootstrapping always
        # gives different result. However, other properties could be tested
        expected_shifts = [randint(2, 9) * 2 for _ in range(3)]
        df = make_df_from_expected_shifts(expected_shifts)
        shifted_test = ShiftedLinearCoefficient(
            target_col="A",
            max_shift=5,
            bootstrap_iterations=500,
            bootstrap_samples=1000,
        )
        shifted_test.fit(df)

        linear_p_values = shifted_test.p_values_
        for col_index in range(len(linear_p_values.columns)):
            assert linear_p_values.iloc[col_index, col_index] == 0

    def test_pearson_p_values(self):
        expected_shifts = [randint(2, 9) * 2 for _ in range(3)]
        df = make_df_from_expected_shifts(expected_shifts)
        shifted_test = ShiftedPearsonCorrelation(
            target_col="A",
            max_shift=5,
            bootstrap_iterations=500,
            bootstrap_samples=1000,
        )
        shifted_test.fit(df)

        pearson_p_values = shifted_test.p_values_
        for col_index in range(len(pearson_p_values.columns)):
            assert pearson_p_values.iloc[col_index, col_index] == 0


def make_df_from_expected_shifts(expected_shifts: List[int]) -> pd.DataFrame:
    testing.N, testing.K = 500, 1

    df = testing.makeTimeDataFrame(freq="D")
    for sh, k in zip(expected_shifts, range(3)):
        df[f"shift_{k}"] = df["A"].shift(sh)
    df = df.dropna()

    return df


def shift_df_from_expected_shifts(
    df: pd.DataFrame, expected_shifts: List[int]
) -> pd.DataFrame:
    for sh, k in zip(expected_shifts, range(3)):
        df[f"shift_{k}"] = df[f"shift_{k}"].shift(-sh)
    return df.dropna()
